#include "bruteforce.h"
#include "common.h"
#include <db.h>
#include "scoring.h"
#include "perm.h"
#include "bitstring.h"

/* Adjacency List Management */
anode anode_insert(anode_pool pool,anode head,int value) {
  anode cur,prev,
    me=anode_alloc(pool,value);

  /* Trivial Insert */
  if(head == NULL)
    return me;
  /* Head Insert Special Case */
  if(value < head->value) {
    me->next = head;
    return me;
  }
  prev = cur = head;
  while(cur && cur->value < value) {
    prev = cur;cur = cur->next;
  }
  me->next = cur;
  prev->next = me;
  return head;
}
anode anode_remove(anode_pool pool,anode head,int value) {
  anode cur,prev;
  if(head == NULL) return NULL;	/* Not here! */
  if(head->value == value) {
    cur = head->next;
    anode_free(pool,head);
    return cur;
  }
  prev = cur = head;
  while(cur && cur->value != value) {
    prev = cur;cur = cur->next;
  }
  if(cur != NULL) {
    prev->next = cur->next;
    anode_free(pool,cur);
  }
  return head;
}

void anode_pool_init(anode_pool pool,int n,int m_p) {
  int   i;
  anode a;
  pool->max = pool->fp = n*m_p;
  pool->nodes    = (anode)R_alloc(sizeof(struct anode),pool->max);
  pool->freelist = (int*)R_alloc(sizeof(int),pool->max);
  a = pool->nodes;
  for(i=0;i<pool->max;i++) {pool->freelist[i] = (int)a;a++;}
}

anode anode_alloc(anode_pool pool,int value) {
  if(pool->fp) {
    anode a = (anode)(pool->freelist[--pool->fp]);
    a->next = NULL;
    a->value= value;
    return a;
  } else error("Freelist empty?!");
}
void anode_free(anode_pool pool,anode a) {
  pool->freelist[pool->fp++] = (int)a;
}

anode_graph anode_graph_create(int n,int m_p) {
  anode_graph g = (anode_graph)R_alloc(sizeof(struct anode_graph),1);
  g->n = n;
  g->graph   = (char*)R_alloc(sizeof(char),bitstr_size(n*n));
  g->parents = (int*)R_alloc(sizeof(int),n);
  g->color   = (int*)R_alloc(sizeof(int),n);
  g->order   = (int*)R_alloc(sizeof(int),n);
  g->vertices= (anode*)R_alloc(sizeof(anode),n);
  anode_pool_init(&(g->pool),n,m_p);
  memset(g->graph,0,sizeof(char)*bitstr_size(n*n));
  memset(g->parents,0,sizeof(int)*n);
  memset(g->vertices,0,sizeof(anode)*n);
  return g;
}

void anode_graph_add_edge(anode_graph g,int from,int to) {
  if(!bit_test(g->graph,(to*g->n)+from)) {
    bit_set(g->graph,(to*g->n)+from);
    g->parents[to]++;
    g->vertices[from] = anode_insert(&(g->pool),g->vertices[from],to);
  }
}
void anode_graph_remove_edge(anode_graph g,int from,int to) {
  if(bit_test(g->graph,(to*g->n)+from)){
    bit_clear(g->graph,(to*g->n)+from);
    g->parents[to]--;
    g->vertices[from] = anode_remove(&(g->pool),g->vertices[from],to);
  }
}
void anode_graph_clear_parents(anode_graph g,int to) {
  int from;
  for(from=0;from<g->n;from++) anode_graph_remove_edge(g,from,to);
}


int anode_graph_is_cyclic(anode_graph g) {
  int i,j,sp,op,oq,edges,*stack;

  oq = op = 0;
  memcpy(g->color,g->parents,sizeof(int)*g->n);
  for(j=0;j<g->n;j++) {
    if(g->parents[j] == 0)
      g->order[op++] = j;
  }
  
  while(oq<op) {
    int node = g->order[oq++];
    anode  n = g->vertices[node];
    while(n) {
      if(--g->color[n->value] == 0) g->order[op++] = n->value;
      n = n->next;
    }
  }
  /* Couldn't complete the order, must be a cycle */
  if(op < g->n) return 1;
  return 0;
}

pset_iter pset_iter_create(int vertex,int n,int m_p) {
  pset_iter pi = (pset_iter)R_alloc(sizeof(struct pset_iter),1);
  pi->vertex = vertex;
  pi->n      = n;
  pi->P      = (int*)R_alloc(sizeof(int),pi->n+1);
  pi->p      = 0;
  pi->cont   = 0;
  pi->m_p    = m_p;
  return pi;
}

void pset_iter_reset(pset_iter pi) {
  pi->p    = 0;
  pi->cont = 0;
}

int pset_iter_next(pset_iter pi,anode_graph g) {
  int  j,n=pi->n-1;
  switch(pi->cont) {
  case 0:goto cont_0;
  case 1:goto cont_1;
  case 2:goto cont_2;
  default:return 0;
  }
 cont_0:
  anode_graph_clear_parents(g,pi->vertex);
  pi->cont = 1;
  return 1;
 cont_1:
  pi->p++;
  if(pi->p > pi->m_p) {
    pset_iter_reset(pi);
    return 0;
  }
  anode_graph_clear_parents(g,pi->vertex);
  inittwiddle(pi->p,n,pi->P);
  for(j=0;j!=n-pi->p;j++);
  for(;j!=n;j++) 
    anode_graph_add_edge(g,j<pi->vertex?j:j+1,pi->vertex);
  pi->cont = 2;
  return 1;
 cont_2:
  if(!twiddle(&(pi->x),&(pi->y),&(pi->z),pi->P)) {
    anode_graph_remove_edge(g,(pi->y)<pi->vertex?pi->y:pi->y+1,pi->vertex);
    anode_graph_add_edge(g,pi->x<pi->vertex?pi->x:pi->x+1,pi->vertex);
    return 1;
  } else 
    goto cont_1;
}

void anode_graph_dump(anode_graph g) {
  int i;anode n;
  for(i=0;i<g->n;i++) {
    if(g->vertices[i]) {
      Rprintf("%d-> ",i+1);
      n = g->vertices[i];
      while(n) { Rprintf("%d ",n->value+1);n=n->next; }
      Rprintf("\n");
    }
  }
}

void enumerate_graphs(int n,int m_p,void (*callBack)(void*,anode_graph),
		      void*userData) {
  int j;
  char *vmax = vmaxget();
  anode_graph g = anode_graph_create(n,m_p);
  pset_iter  *i = (pset_iter*)R_alloc(sizeof(pset_iter),n);

  /* Initialize the iterators */
  for(j=0;j<n;j++) {
    i[j] = pset_iter_create(j,n,m_p);
    if(j) pset_iter_next(i[j],g);
  }
  while(1) {
    /* Step the iterators */
    for(j=0;j<n;j++)
      if(pset_iter_next(i[j],g)) break;
    if(j==n) break;		/* All done! */
    if(callBack && !anode_graph_is_cyclic(g))
      callBack(userData,g);
  }
  vmaxset(vmax);
}

/*  10,365,456,954 */

/*   82,923,655,629       */
/*    1,073,741,824 = 1GB */
/* Estimated 10GB of storage space required */
struct dumper {
  int   dag_size;
  char *orig;
  char  filename[512];
  char *buffer,*cur;
  int   n,max;
  int   fileno;
  long  count;
  FILE  *fd;
};

void dump_callback(void*userData,anode_graph g) {
  struct dumper *d = (struct dumper*)userData;
  d->count++;
  memcpy(d->cur,g->graph,d->dag_size);
  if(++d->n < d->max){ 
    d->cur += d->dag_size;
  } else {
    d->n = 0;
    d->cur = d->buffer;
    fwrite(d->buffer,d->max,d->dag_size,d->fd);
    fclose(d->fd);
    Rprintf("%s %d...\n",d->filename,d->count);
    sprintf(d->filename,"%s.%d",d->orig,d->fileno++);
    d->fd = fopen(d->filename,"w+");
  }
}

void count_callback(void*userData,anode_graph g) {
  long*count = (long*)userData;
  count[0]++;
  if(count[0] && count[0] % 1000000 == 0) Rprintf("%d...",count[0]);
}

SEXP BNSL_EnumerateGraphs(SEXP n,
			  SEXP max_parents,
			  SEXP file) {
  struct dumper d;
  d.orig   = CHAR(STRING_ELT(file,0));
  d.fileno = 0;
  sprintf(d.filename,"%s.%d",d.orig,d.fileno++);
  d.fd = fopen(d.filename,"w+");
  d.dag_size = bitstr_size(INTEGER(n)[0]*INTEGER(n)[0]);
  d.max      = 100000000;
  d.cur = d.buffer   = (char*)R_alloc(d.dag_size,d.max);
  d.count = d.n   = 0;
  enumerate_graphs(INTEGER(n)[0],
		   INTEGER(max_parents)[0],
		   dump_callback,&d);
  if(d.n > 0)
    fwrite(d.buffer,d.n,d.dag_size,d.fd);
  fclose(d.fd);
  return ScalarReal(d.count);
}

