#include <bitstring.h>
#include "common.h"
#include <Rdefines.h>
#include <db.h>
#include "scoring.h"
#include "dag.h"

SEXP edge_list(int n) {
  SEXP list,edges;
  PROTECT(list = allocVector(VECSXP,1));
  if(n>0) { 
    PROTECT(edges = allocVector(INTSXP,n));
    SET_VECTOR_ELT(list,0,edges);
    UNPROTECT(1);
  } else SET_VECTOR_ELT(list,0,R_NilValue);
  setAttrib(list,R_NamesSymbol,ScalarString(mkChar("edges")));
  UNPROTECT(1);
  return list;
}


SEXP dag_to_graph(dag_ctx ctx,const char* klass_name) {
  int  i,j;
  SEXP G,names,edgeL;
  char nodeName[64];
  PROTECT(G     = NEW_OBJECT(MAKE_CLASS((char*)klass_name)));
  PROTECT(names = allocVector(STRSXP,ctx->n));
  PROTECT(edgeL = allocVector(VECSXP,ctx->n));
  for(i=0;i<ctx->n;i++) {
    SEXP eL,edges;
    sprintf(nodeName,"%d",i+1);
    SET_STRING_ELT(names,i,mkChar(nodeName));
    PROTECT(eL = edge_list(bnsl_dag_degree(ctx,i+1)));
    SET_VECTOR_ELT(edgeL,i,eL);
    edges = VECTOR_ELT(eL,0);
    if(edges != R_NilValue) {
      int r=0;
      for(j=0;j<ctx->n;j++) 
	if(bnsl_dag_has_edge(ctx,i+1,j+1)) INTEGER(edges)[r++] = j+1;
    }
    UNPROTECT(1);
  }
  setAttrib(edgeL,R_NamesSymbol,duplicate(names));
  SET_SLOT(G,install("edgeL"),   edgeL);
  SET_SLOT(G,install("nodes"),   names);
  SET_SLOT(G,install("edgemode"),ScalarString(mkChar("directed")));
  UNPROTECT(3);
  return G;
}

void bnsl_randomize_graph(dag_ctx ctx,int*chain) {
  int i,j;
  int steps = 4*(ctx->n*ctx->n);
  for(i=0;i<ctx->n-1;i++)
    bnsl_dag_add_edge(ctx,chain[i]+1,chain[i+1]+1);
  GetRNGstate();
  for(i=0;i<steps;i++) {
  }
  PutRNGstate();
}

SEXP BNSL_GibbsDAG(SEXP D,SEXP K,SEXP STEPS) {
  int i,j,a,b;
  dag_ctx ctx = bnsl_dag_ctx(D,INTEGER(K)[0]);
  int*    ndx = (int*)R_alloc(sizeof(int),ctx->n);
  int*    ndx2= (int*)R_alloc(sizeof(int),ctx->n);
  for(i=0;i<ctx->n;i++) {
    ndx[i] = i; /* Initialize the ordering */
    ndx2[i]= i;
    bnsl_dag_update_score(ctx,i+1);
  }
/*   /\* Initialize a chain graph *\/ */
/*   for(a=0;a<ctx->n-1;a++) { */
/*     int b = ((int)(((double)(ctx->n-a))*unif_rand())); */
/*     int t = ndx[a]; */
/*     ndx[a] = ndx[b]; */
/*     ndx[b] = t; */
/*   } */
/*   bnsl_randomize_graph(dag_ctx,ndx); */

  Rprintf("empty network score: %.2f\n",bnsl_dag_score(ctx));

  GetRNGstate();
  for(i=0;i<INTEGER(STEPS)[0];i++) {
    /* Shuffle the ordering */
    for(a=0;a<ctx->n-1;a++) {
      int b = ((int)(((double)(ctx->n-a))*unif_rand()));
      int t = ndx[a];
      ndx[a] = ndx[b];
      ndx[b] = t;
    }
    /* Shuffle the ordering of ndx2 */
    for(a=0;a<ctx->n-1;a++) {
      int b = ((int)(((double)(ctx->n-a))*unif_rand()));
      int t = ndx2[a];
      ndx2[a] = ndx2[b];
      ndx2[b] = t;
    }


    for(a=0;a<ctx->n;a++) {
      for(b=0;b<ctx->n;b++) {
	if(ndx[a] != ndx2[b]) {
	  int edge;
	  double current,update;
	  current = bnsl_dag_get_score(ctx,ndx2[b]+1);
	  edge    = bnsl_dag_has_edge(ctx,ndx[a]+1,ndx2[b]+1);
	  if(!edge) {
	    if(bnsl_dag_degree(ctx,ndx2[b]+1) < INTEGER(K)[0] &&
	       bnsl_dag_add_edge(ctx,ndx[a]+1,ndx2[b]+1)) {
	      double Premove;
	      bnsl_dag_update_score(ctx,ndx2[b]+1);
	      update = bnsl_dag_get_score(ctx,ndx2[b]+1);
	      Premove = 1.0/(1.0+exp(update-current));
/* 	      Rprintf("%.2f %.2f P(%d->%d) == %.2f\n",current,update,ndx[a]+1,ndx2[b]+1,1.0-Premove); */
	      if(unif_rand() < Premove) {
		bnsl_dag_remove_edge_nocheck(ctx,ndx[a]+1,ndx2[b]+1);
		bnsl_dag_update_score(ctx,ndx2[b]+1);
	      }
	    }/*  else  */
/* 	      Rprintf("P(%d->%d) == 0\n",ndx[a]+1,ndx2[b]+1); */
	  } else {
	    double Premove;
	    bnsl_dag_remove_edge_nocheck(ctx,ndx[a]+1,ndx2[b]+1);
	    bnsl_dag_update_score(ctx,ndx2[b]+1);
	    update = bnsl_dag_get_score(ctx,ndx2[b]+1);
	    Premove = 1.0/(1.0+exp(current-update));
/* 	    Rprintf("%.2f %.2f P(%d->%d) == %.2f\n",current,update,ndx[a]+1,ndx2[b]+1,1.0-Premove); */
	    if(unif_rand() > Premove) {
	      bnsl_dag_add_edge(ctx,ndx[a]+1,ndx2[b]+1);
	      bnsl_dag_update_score(ctx,ndx2[b]+1);
	    }
	  }
	}
      }
    }
    /*    bnsl_dag_dump(ctx); */
    Rprintf("%.2f\n",bnsl_dag_score(ctx));
  }
  bnsl_dag_ctx_done(ctx);
  PutRNGstate();
  return R_NilValue;
}

SEXP BNSL_SearchForDAG(SEXP D,SEXP K,SEXP EL,SEXP STEPS) {
  int i,j;
  SEXP retList = R_NilValue;
  dag_ctx ctx = bnsl_dag_ctx(D,INTEGER(K)[0]);
  for(i=0;i<length(EL);i+=2) {
    if(!bnsl_dag_add_edge(ctx,INTEGER(EL)[i],INTEGER(EL)[i+1]))
      error("Unable to add edge, likely it induces a cycle in the DAG");
    bnsl_dag_update_score(ctx,INTEGER(EL)[i+1]);
  }
  Rprintf("Starting from %.4f\n",bnsl_dag_score(ctx));

  /* Run the sampler for a while to get a random DAG */
  GetRNGstate();
  for(j=0;j<10;j++) {
    int    accept  = 0;
    double current = bnsl_dag_score(ctx);
    for(i=0;i<INTEGER(STEPS)[0];i++) {
      int a,b;
      double proposal = current;
      int    undo_code= 0;
      a = (int)((double)ctx->n*unif_rand())+1;
      do { b = (int)((double)ctx->n*unif_rand())+1; } while(a == b);
      
      /* If we have the arc */
      if(bnsl_dag_has_edge(ctx,a,b)) {
	/* Remove the edge or reverse it if it is disconnecting */
	if(!bnsl_dag_remove_edge(ctx,a,b)) {
	  if(bnsl_dag_degree(ctx,a) < INTEGER(K)[0]) {
	    bnsl_dag_reverse_edge(ctx,a,b);
	    bnsl_dag_update_score(ctx,b);
	    bnsl_dag_update_score(ctx,a);
	    proposal = bnsl_dag_score(ctx);
	    undo_code = 1;	/* reverse edge */
	  }
	} else {
	  undo_code=2;		/* remove edge */
	  bnsl_dag_update_score(ctx,b);
	  proposal = bnsl_dag_score(ctx);
	}
      } else {
	/* Add an edge or fail (if it induces a cycle) */
	if(bnsl_dag_degree(ctx,b) < INTEGER(K)[0])
	  if(bnsl_dag_add_edge(ctx,a,b)) {
	    bnsl_dag_update_score(ctx,b);
	    proposal = bnsl_dag_score(ctx);
	    undo_code = 3; 	/* add edge */
	  }
      }
      if(unif_rand() > exp(proposal-current)) {
	switch(undo_code) {
	case 1:
	    bnsl_dag_reverse_edge(ctx,b,a);
	    bnsl_dag_update_score(ctx,b);
	    bnsl_dag_update_score(ctx,a);
	    break;
	case 2:
	  if(!bnsl_dag_add_edge(ctx,a,b))
	    bnsl_dag_dump(ctx);
	  bnsl_dag_update_score(ctx,b);
	  break;
	case 3:
	  if(!bnsl_dag_remove_edge(ctx,a,b))
	    bnsl_dag_dump(ctx);
	  bnsl_dag_update_score(ctx,b);
	  break;
	}
	if(current != bnsl_dag_score(ctx)) {
	  Rprintf("While attempting to undo %d, got an inconsistent return\n",
		  undo_code);
	  bnsl_dag_dump(ctx);
	}
	Rprintf("(%d [%d %d]) %.2f %.4f %.4f\n",
		undo_code,a,b,
		exp(proposal-current),current,bnsl_dag_score(ctx));
      } else {
	/* Staying in place isn't accepting. */
	if(current != proposal) accept++;
	Rprintf("(%d %c) %.2f %.4f\n",
		undo_code,(current!=proposal)?'+':'-',
		((double)accept)/((double)(i+1)),proposal);
	current = proposal;
      }
    }
    bnsl_dag_dump(ctx);
    Rprintf("%.4f %.4f %.2f---\n",
	    current,bnsl_dag_score(ctx),
	    ((double)accept)/((double)INTEGER(STEPS)[0]));
  }
  PutRNGstate();
  bnsl_dag_ctx_done(ctx);
  /* Recalculate and return the edge list for this DAG */
  return retList;
}


SEXP BNSL_RawToDag(SEXP R,SEXP N) {
  int n = INTEGER(N)[0];
  int s = bitstr_size(n);
  int i,j;
  Rbyte *f;double *d;
  SEXP mat,dim;
  PROTECT(mat = allocVector(REALSXP,n*n));
  PROTECT(dim = allocVector(REALSXP,2));
  REAL(dim)[0] = (double)n;
  REAL(dim)[1] = (double)n;
  f = RAW(R);d = REAL(mat);
  memset(d,0,sizeof(double)*n*n);
  for(i=0;i<n;i++) {
    for(j=0;j<n;j++) if(bit_test(f,j)) d[j] = 1.0;
    f += s;
    d += n;
  }
  setAttrib(mat,R_DimSymbol,dim);
  UNPROTECT(2);
  return mat;
}
