#include <Rinternals.h>
#include <db.h>
#include "dag.h"
#include "bitstring.h"


DagState DagStateInit(int n,int k,int kk) {
  DagState ds = (DagState)R_alloc(sizeof(struct DagState),1);
  ds->n = n;ds->k = k;ds->kk = kk;

  ds->adjM    = (char*)R_alloc(sizeof(char),bitstr_size(n)*n);
  memset(ds->adjM,0,sizeof(char)*bitstr_size(n)*n);
  ds->parent  = (char*)R_alloc(sizeof(char),n*n);
  memset(ds->parent,0,sizeof(char)*n*n);
  ds->child   = (char*)R_alloc(sizeof(char),n*n);
  memset(ds->child,0,sizeof(char)*n*n);

  ds->stack   = (char*)R_alloc(sizeof(char),n*n);
  ds->color   = (char*)R_alloc(sizeof(char),n);
  memset(ds->stack,0,sizeof(char)*n*n);
  memset(ds->color,0,sizeof(char)*n);

  return ds;
}

#define ADJ(A,B) bit_test(ds->adjM+bitstr_size(ds->n)*a,b)

/* Check to see if we can add this edge to an acyclic network */
int DagStateProposedCheck(DagState ds,int a,int b) {
  int sp = 0,i = 0;
  memset(ds->color,0,sizeof(char)*ds->n);
  ds->color[a]    = 1;
  ds->stack[sp++] = a;
  while(i < sp) {
    int node = ds->stack[i++]; 	        /* Shift off the stack */
    char  *q = ds->parent + ds->n*node; /* Get parent list */
    int    j;
    /* Move new parents onto the stack */
    for(j=0;j<ds->fanIn[node];j++) 
      if(q[j] == b) return 0; else if(!ds->color[q[j]]) { ds->stack[sp++] = q[j];ds->color[q[j]]=1; }
  }
  return 1;
}

int DagStateHasEdge(DagState ds,int a,int b) {
  return bit_test(ds->adjM + a*bitstr_size(ds->n),b);
}

int DagStateAdd(DagState ds,int a,int b,int cycleCheck) {
  char *q;
  if(ds->adjM[ds->n*a + b])                            return 0;
  /* Reject fanIn and fanOut violations */
  if(ds->fanIn[b] >= ds->k || ds->fanOut[a] >= ds->kk) return 0;
  /* Do our cycle checking if requested */
  if(cycleCheck && !DagStateProposedCheck(ds,a,b))     return 0;

  /* Update the parent and child lists */
  bit_set(ds->adjM + a*bitstr_size(ds->n),b);
  q = ds->parent + ds->n*b;
  q[ds->fanIn[b]++] = a;
  q = ds->child  + ds->n*a;
  q[ds->fanOut[a]++] = b;

  return 1;
}

int DagStateConnectedPropose(DagState ds,int a,int b) {
  int sp = 0,i = 0,j;
  char *q;
  memset(ds->color,0,sizeof(char)*ds->n);
  ds->color[0] = 1;
  /* Initialize Stack */
  for(j=0;j<ds->fanIn[0];j++) 
    if(b !=0 || ds->parent[j] != a) ds->stack[sp++] = ds->parent[j];
  for(j=0;j<ds->fanOut[0];j++) 
    if(a != 0 || ds->child[j] != b) ds->stack[sp++] = ds->child[j];
  while(sp) {
    int node = ds->stack[--sp];

    ds->color[node] = 1;
    char  *q = ds->parent + ds->n*node;
    for(j=0;j<ds->fanIn[node];j++) 
      if(ds->color[q[j]] == 0 && (b != node || q[j] != a)) ds->stack[sp++] = q[j];
    q = ds->child + ds->n*node;
    for(j=0;j<ds->fanOut[node];j++) 
      if(ds->color[q[j]] == 0 && (a != node || q[j] != b)) ds->stack[sp++] = q[j];
  }
  /* Ensure visitation of all nodes */
  for(j=0;j<ds->n;j++) if(ds->color[j] == 0) return 0;
  return 1;
}

char *DagStateParentSet(DagState ds,int a) {
  return ds->adjM + bitstr_size(ds->n)*a;
}

int DagStateDel(DagState ds,int a,int b,int connectCheck) {
  char *q;
  int   j;
  if(ds->adjM[ds->n*a + b] == 0)                        return 0;
  if(connectCheck && !DagStateConnectedPropose(ds,a,b)) return 0;
  bit_clear(ds->adjM + bitstr_size(ds->n)*a,b);
  
  q = ds->parent + ds->n*b;
  for(j=0;j<ds->fanIn[b];j++) q[j] = a;
  if(j < ds->fanIn[b]) q[j] = q[--ds->fanIn[b]]; else --ds->fanIn[b];
  q = ds->child  + ds->n*a;
  for(j=0;j<ds->fanOut[a];j++) q[j] = b;
  if(j < ds->fanOut[a]) q[j] = q[--ds->fanOut[a]]; else --ds->fanOut[a];
  return 1;
}

void DagStateDump(DagState ds) {
  int i;
  Rprintf("<<\n");
  for(i=0;i<ds->n;i++) {
    int j;char *q;
    Rprintf("\t(");
    q = ds->parent + ds->n*i;
    for(j=0;j<ds->fanIn[j];j++) Rprintf("%d ",q[j]+1);
    Rprintf(") %d (",i+1);
    q = ds->child + ds->n*i;
    for(j=0;j<ds->fanOut[j];j++) Rprintf("%d ",q[j]+1);
    Rprintf(")\n");
  }
  Rprintf(">>\n");
}
