#include <Rdefines.h>
#include "common.h"
#include "bitstring.h"
#include <db.h>
#include <Rmath.h>
#include "fastdb.h"
#include "fastdag.h"

void add_pset(pset** psets,int *p,int n,int k,int i,
	      double score) {
  pset *ps = (pset*)R_alloc(sizeof(struct _pset)+bitstr_size(n),1);
  int   j;
  /* Special case here */
  if(p[0] == 0) k=0;
  memset(ps,0,sizeof(struct _pset)+k+bitstr_size(n));
  char *PS = ps->pset + k;
  ps->n          = k;
  ps->score      = score;
  ps->real_score = score;	/* Store an unnormalized score */
  for(j=0;j<k;j++) {
    ps->pset[j] = p[j]-1;
    bit_set(PS,p[j]-1);
  }

  // Insert into linked list
  if(psets[i] == NULL || psets[i]->score < score) {
    ps->next = psets[i];
    psets[i] = ps;
  } else {
    pset *n = psets[i];
    while(n->next != NULL) {
      if(n->next->score <= score) break;
      n = n->next;
    }
    ps->next = n->next;
    n->next  = ps;
  }
}

void FillInversions(fast_score_ctx ctx,DB*over,SEXP orders,SEXP inversions) {

  int *overlap = (int*)R_alloc(sizeof(int),ctx->n);
  DBT data;
  memset(&data,0,sizeof(DBT));
  data.size = sizeof(int)*ctx->n;
  data.data = overlap;
  int i,j,n,ret;
  for(ret = ctx->cur->c_get(ctx->cur,&(ctx->key),&(ctx->data),DB_FIRST);
      ret == 0;
      ret = ctx->cur->c_get(ctx->cur,&(ctx->key),&(ctx->data),DB_NEXT)) {
    int     k = ctx->key.size/sizeof(int);
    int    *p = (int*)(ctx->key.data);
    double *s = (double*)(ctx->data.data);
    double  b2 = ctx->gamma*((double)k);
    int     m = ctx->data.size/sizeof(double);
    if(k == 0 || k > ctx->K) continue;
    memset(overlap,0,sizeof(int)*ctx->n);
    
    if(p[0] == 0)
      for(i=0;i<n;i++) 
	overlap[i] = length(inversions);
    else {
      for(j=0;j<length(inversions);j++) {
	SEXP inv = VECTOR_ELT(inversions,j);
	SEXP ord = VECTOR_ELT(orders,j);
	int q,max = 0;
	for(q=0;q<k;k++) if(max < INTEGER(inversions)[q]) max = INTEGER(inversions)[q];
	for(q=max+1;q<length(ord);q++) {
	  overlap[INTEGER(ord)[q]-1]++;
	}
      }
    }
    over->put(over,NULL,&(ctx->key),&data,0);
  }
}


fast_dag_ctx fast_dag(fast_score_ctx ctx,int *o,int n) {
  fast_dag_ctx fd = (fast_dag_ctx)R_alloc(sizeof(struct fast_dag_ctx)+
					  sizeof(pset*)*ctx->n,1);
  fd->sizes = (int*)R_alloc(sizeof(int),ctx->n);
  fd->n     = ctx->n;
  fd->K     = ctx->K;
  memset(fd->psets,0,sizeof(pset*)*ctx->n);
  memset(fd->sizes,0,sizeof(int)*ctx->n);
  int ret,i,j;
  for(i=0;i<n;i++) {
    ctx->inv[o[i]-1] = i;
    ctx->score[i] = R_NegInf;
  }
  for(ret = ctx->cur->c_get(ctx->cur,&(ctx->key),&(ctx->data),DB_FIRST);
      ret == 0;
      ret = ctx->cur->c_get(ctx->cur,&(ctx->key),&(ctx->data),DB_NEXT)) {
    int     k = ctx->key.size/sizeof(int);
    int    *p = (int*)(ctx->key.data);
    double *s = (double*)(ctx->data.data);
    double  b2 = -ctx->gamma*((double)k);
    int     m = ctx->data.size/sizeof(double);
    if(k == 0 || k > ctx->K) continue;
    /*    for(i=0;i<n;i++) ctx->score; */
    if(p[0] == 0)
      for(i=0;i<n;i++) {
	//	b1 = ctx->gamma*ctx->log_opp_b*
	//	  ((ctx->inv[o[i]-1]<ctx->K)?ctx->inv[o[i]-1]:ctx->K);
	add_pset(fd->psets,p,ctx->n,k,o[i]-1,s[o[i]-1]);
	fd->sizes[o[i]-1]++;
	ctx->score[o[i]-1] = logspace_add(ctx->score[o[i]-1],s[o[i]-1]);
      }
    else {
      i=0;
      for(j=0;j<k;j++) if(i <= ctx->inv[p[j]-1]) i = ctx->inv[p[j]-1]+1;
      for(;i<n;i++) {
	//	b1 = ctx->log_opp_b*
	//	  ((ctx->inv[o[i]-1]<ctx->K)?ctx->inv[o[i]-1]:ctx->K);
	add_pset(fd->psets,p,ctx->n,k,o[i]-1,s[o[i]-1]+b2);
	fd->sizes[o[i]-1]++;
	ctx->score[o[i]-1] = logspace_add(ctx->score[o[i]-1],
					  s[o[i]-1]+b2);
      }
    }
  }
  //Normalize everybody
  for(i=0;i<ctx->n;i++) {
    pset*p = fd->psets[i];
    while(p != NULL) {
      p->score = exp(p->score-ctx->score[i]);
      p = p->next;
    }
  }

  return fd;
}

void fast_dag_sample(fast_dag_ctx ctx,int N,
		     void callback(void*,char*,int,double),void*user) {
  int   i,size = bitstr_size(ctx->n)*ctx->n;
  char *dag  = R_alloc(sizeof(char),size);
  for(i=0;i<N;i++) {
    int j;
    double os = 1.0;
    double us = 0.0;
    for(j=0;j<ctx->n;j++) {
      double u = unif_rand();
      double c = 0.0;
      pset*  p = ctx->psets[j];
      if(!p) error("Unable to sample dag, no parent sets");
      //      Rprintf("(%.2f) ",u);
      while(p) { 
	//	Rprintf("%.2f ",p->score);
	c += p->score;
	if(u <= c) break;
	p = p->next; 
      }
      //      Rprintf("[%.2f]\n",c);
      os *= p->score;
      us += p->real_score;
      memcpy(&dag[bitstr_size(ctx->n)*j],p->pset+p->n,bitstr_size(ctx->n));
    }
    callback(user,dag,size,us);
  }
}

struct dag_info {
  DB*  db;
  DBT  key,data;
  int  n;
  SEXP orders;
  SEXP inversion;
  int  unique;
};

struct _x {
  int    count;
  double score;
};

extern int IsConsistent(int n,char*rep,int*order,char*topo);

void insert_dag(void* userInfo,char*dag,int size,double score) {
  struct dag_info *di = (struct dag_info*)userInfo;
  di->key.data  = dag;
  di->key.size  = size;
  if(di->db->get(di->db,NULL,&(di->key),&(di->data),0) == 0) {
    struct _x *x = di->data.data;
    x->count++;
  } else {
    int i;
    di->unique++;
    struct _x x;
    x.count = 1;
    x.score = score;
    di->data.data = &x;
    di->data.size = sizeof(struct _x);
  }
  di->db->put(di->db,NULL,&(di->key),&(di->data),0);
}

void init_dag_info(struct dag_info *di,int n,SEXP orders) {

  memset(di,0,sizeof(struct dag_info));
  db_create(&(di->db),NULL,0);
  di->db->open(di->db,NULL,NULL,NULL,DB_BTREE,DB_CREATE,0);
  di->n = n;
  di->orders = orders;
}

pset *get_pset(pset*head,int n) {
  while(n-- > 0) head = head->next;
  return head;
}

int check_all_but(SEXP inv,pset **ps,int n,int *ndx,int but) {
  int i,count = 0;
  for(i=0;i<length(inv);i++) {
    SEXP o = VECTOR_ELT(inv,i);
    int j,l,cancel = 0;
    for(j=0;j<n;j++) {
      if(j == but) continue;
      for(l=0;l<ps[j]->n;l++) if(INTEGER(o)[ps[j]->pset[l]] > INTEGER(o)[j]) { cancel = 1;break; }
    }
    if(!cancel) {
      ndx[count++] = i;
    }
  }
  return count;
}

extern int unif_int(int);

SEXP BNSL_FastCorrected(SEXP D,SEXP K,SEXP os) {
  fast_score_ctx ctx = fast_ctx(D,INTEGER(K)[0]);
  SEXP orders = GET_SLOT(os,install("orders"));
  SEXP counts = GET_SLOT(os,install("counts"));
  GetRNGstate();
  int unique,COUNT;
  DB  *db;
  DBT  key,data;
  char *rawdag;
  db_create(&db,NULL,0);
  db->open(db,NULL,NULL,NULL,DB_BTREE,DB_CREATE,0);
  rawdag = (char*)R_alloc(sizeof(char)*bitstr_size(ctx->n),ctx->n);
  memset(&key,0,sizeof(DBT));
  memset(&data,0,sizeof(DBT));

  unique = 0;

  /* Invert all orders */
  SEXP ivert;
  Rprintf("Inverting Orders...");
  PROTECT(ivert = allocVector(VECSXP,length(orders)));
  int i;
  for(i=0;i<length(orders);i++) {
    SEXP o = VECTOR_ELT(orders,i);
    SEXP inv;
    PROTECT(inv = allocVector(INTSXP,length(o)));
    int j;
    for(j=0;j<length(o);j++) INTEGER(inv)[INTEGER(o)[j]-1] = j;
    SET_VECTOR_ELT(ivert,i,inv);
    UNPROTECT(1);
  }
  Rprintf("Done\n");

  int   *ndx = (int*)R_alloc(sizeof(int),length(orders));
  pset **cur = (pset**)R_alloc(sizeof(pset*),ctx->n);
  /* Scan Order */
  int   *so  = (int*)R_alloc(sizeof(int),ctx->n);
  for(i=0;i<ctx->n;i++)
    so[i] = i;
  
  for(i=0;i<length(orders);i++) {
    SEXP order = VECTOR_ELT(orders,i);
    char *vmax = vmaxget();
    fast_dag_ctx dag = fast_dag(ctx,INTEGER(order),length(order));
    int j;
    int iter = INTEGER(counts)[i];
    if(iter < 100) iter=100;
    Rprintf("%d of %d (%d)...",i+1,length(orders),iter);
    /* Initialize a starting location */
    for(j=0;j<ctx->n;j++)
      cur[j] = get_pset(dag->psets[j],unif_int(dag->sizes[j]));
    /* Sample DAGs from this ordering */
    for(j=0;j<2*iter;j++) {
      /* A single scan o' the gibbs sampler */
      int c;
      for(c=0;c<ctx->n;c++) {
	int a = c+unif_int(ctx->n-c);
	int t = so[a];
	so[a] = so[c];
	so[c] = t;
	if(dag->sizes[t] == 0) continue;	/* We don't do anything for the first entry */
	/* Update the overlap set */
	int pos = check_all_but(ivert,cur,ctx->n,ndx,t);
	/* Update the scores, normalize and sample */
	/*	Rprintf("%d orders selected\n",pos);*/
	/* Recalculate scores for parent sets and renormalize */
	pset *ps = dag->psets[t];

	/*	Rprintf("%d (%d)>>",t+1,dag->sizes[t]);*/
	double max = R_PosInf;
	for(;ps;ps=ps->next) {
	  int count=0;
	  for(a=0;a<pos;a++) {
	    SEXP inv = VECTOR_ELT(ivert,ndx[a]);
	    int b,y=1;
	    for(b=0;b<ps->n;b++) if(INTEGER(inv)[t] < INTEGER(inv)[ps->pset[b]]) { y=0;break; }
	    if(y) count++;
	  }
	  ps->score = ps->real_score - log(count);
	  /*	  Rprintf("%.2f->%.2f(%d) ",ps->real_score,ps->score,count);*/
	  if(max == R_PosInf || ps->score > max) max = ps->score;
	}
	ps = dag->psets[t];
	double norm = 0.0;
	for(;ps;ps = ps->next) 
	  norm += exp(ps->score-max);
	norm = log(norm)+max;
	/*	Rprintf(" Normalizing Constant: %.2f\n",norm);*/
	
	/* Resample! */
	double cdf = 0.0;
	double u   = unif_rand();
	ps = dag->psets[t];
	/*	Rprintf("%d: ",t+1); */
	for(;ps;ps = ps->next) {
	  cdf += exp(ps->score-norm);
	  /*	  Rprintf("%.2f ",cdf);*/
	  if(u < cdf) break;
	}
	/*	Rprintf("\n");*/
	cur[t] = ps;
      } /* End Scan */
      /* Record after burnin */
      if(j>iter) {
	int a;
	for(a = 0;a < ctx->n;a++)
	  memcpy(rawdag + a*bitstr_size(ctx->n),cur[a]->pset+cur[a]->n,bitstr_size(ctx->n));
	key.data = rawdag;
	key.size = bitstr_size(ctx->n)*ctx->n;
	if(db->get(db,NULL,&key,&data,0) == 0)
	  ((int*)data.data)[0]++;
	else {
	  COUNT = 1;unique++;
	  data.data = &COUNT;
	  data.size = sizeof(int);
	}
	db->put(db,NULL,&key,&data,0);
      }
    }
    Rprintf("%d\n",unique);
    vmaxset(vmax);
  }
  Rprintf("%d unique dags sampled\n",unique);

  DBC *cursor;
  db->cursor(db,NULL,&cursor,0);
  memset(&key,0,sizeof(key));
  memset(&data,0,sizeof(key));
  SEXP ret;
  PROTECT(ret = NEW_OBJECT(MAKE_CLASS("objectSet")));
  SEXP obj,cnt;
  PROTECT(obj = allocVector(VECSXP,unique));
  SET_SLOT(ret,install("objects"),obj);
  PROTECT(cnt = allocVector(INTSXP,unique));
  SET_SLOT(ret,install("counts"),cnt);
  i=0;
  while(cursor->c_get(cursor,&key,&data,DB_NEXT)==0) {
    char *dag  =  key.data;
    int  *count = (int*)data.data;
    SEXP r;
    PROTECT(r = allocVector(RAWSXP,key.size));
    memcpy(RAW(r),dag,key.size);
    SET_VECTOR_ELT(obj,i,r);
    INTEGER(cnt)[i] = count[0];
    UNPROTECT(1);
    i++;
  }
  cursor->c_close(cursor);
  UNPROTECT(3);
  db->close(db,0);
  UNPROTECT(1);
  PutRNGstate();
  return ret;
}

SEXP BNSL_FastDagSample(SEXP D,SEXP os,SEXP K,SEXP N) {
  char          *vmax= vmaxget();
  fast_score_ctx ctx = fast_ctx(D,INTEGER(K)[0]);
  SEXP orders = GET_SLOT(os,install("orders"));
  SEXP counts = GET_SLOT(os,install("counts"));
  DB  *over;
  struct dag_info di;
  init_dag_info(&di,ctx->n,orders);
  /*  PROTECT(di.inversion = allocVector(VECSXP,length(orders)));
  int i;
  for(i=0;i<length(orders);i++) {
    SEXP order = VECTOR_ELT(orders,i);
    SEXP inv;
    PROTECT(inv = allocVector(INTSXP,length(order)));
    int j;
    for(j=0;j<length(order);j++) INTEGER(inv)[INTEGER(order)[i]-1] = j;
    SET_VECTOR_ELT(di.inversion,i,inv);
    UNPROTECT(1);
  }
  db_create(&over,NULL,0);
  over->open(over,NULL,NULL,NULL,DB_BTREE,DB_CREATE,0);
  FillInversions(ctx,over,di.orders,di.inversion);
  UNPROTECT(1);*/
  
  GetRNGstate();
  
  Rprintf("%d unique orders\n",length(orders));
  int i;
  for(i=0;i<length(orders);i++) {
    SEXP order = VECTOR_ELT(orders,i);
    fast_dag_ctx dag = fast_dag(ctx,INTEGER(order),length(order));
    //    Rprintf("sampling %d: ",INTEGER(counts)[0]*INTEGER(N)[0]);
    fast_dag_sample(dag,INTEGER(counts)[0]*INTEGER(N)[0],insert_dag,&di);
    //    Rprintf("%d unique dags\n",di.unique);
    Rprintf(".");
  }
  Rprintf("\n%d unique dags\n",di.unique);

  {
    SEXP ret,scores,dags,sampled,overlap,vertices;
    PROTECT(ret = NEW_OBJECT(MAKE_CLASS("fastDagSample")));
    PROTECT(scores  = allocVector(REALSXP,di.unique));
    PROTECT(dags    = allocVector(VECSXP,di.unique));
    PROTECT(sampled = allocVector(REALSXP,di.unique));
    PROTECT(overlap = allocVector(REALSXP,di.unique));
    PROTECT(vertices= allocVector(REALSXP,di.unique));
    
    DBC *cur;
    memset(&(di.key),0,sizeof(DBT));
    memset(&(di.data),0,sizeof(DBT));
    di.db->cursor(di.db,NULL,&cur,0);
    i=0;
    while(cur->c_get(cur,&(di.key),&(di.data),DB_NEXT)==0) {
      char      *r = di.key.data;
      struct _x *x = di.data.data;
      char      *c = (char*)R_alloc(sizeof(char),ctx->n);
      int        j,tmp;
      SEXP raw;

      PROTECT(raw = allocVector(RAWSXP,di.key.size));
      memcpy(RAW(raw),r,di.key.size);
      SET_VECTOR_ELT(dags,i,raw);
      UNPROTECT(1);
      
      REAL(sampled)[i] = (double)x->count;
      REAL(scores)[i]  = (double)x->score;

      tmp = 0;
      for(j=0;j<length(orders);j++)
	if(IsConsistent(ctx->n,r,INTEGER(VECTOR_ELT(orders,j)),c)) tmp++;
      REAL(overlap)[i] = (double)tmp;

      tmp = 0;
      for(j=0;j<di.key.size;j++) {
	char f = r[j];
	while(f) {
	  tmp++;
	  f &= (f-1);
	}
      }
      REAL(vertices)[i] = (double)tmp;
      i++;
    }
    cur->c_close(cur);

    SET_SLOT(ret,install("dags"),dags);
    SET_SLOT(ret,install("p"),scores);
    SET_SLOT(ret,install("n"),sampled);
    SET_SLOT(ret,install("o"),overlap);
    SET_SLOT(ret,install("v"),vertices);
    UNPROTECT(6);
    di.db->close(di.db,0);
    PutRNGstate();
    vmaxset(vmax);
    return ret;
  }
}
