#include <Rdefines.h>
#include "common.h"
#include "bitstring.h"
#include <db.h>
#include <Rmath.h>
#include "fastdb.h"
#include "eesample.h"

#define Max(X,Y) (((X)>(Y))?(X):(Y))

#define LogTargetRatio(Hnew,Hcur,Hi,T) \
  (Max((Hcur),(Hi))-Max((Hnew),(Hi)))/(T)

extern SEXP OrderRandom(int);
extern SEXP OrderMetropolisProposal(SEXP,int,int*,int*);
extern SEXP OrderMetropolisSwap(SEXP);
extern SEXP OrderMetropolisCut(SEXP);
extern SEXP OrderMetropolisSwapProb(SEXP,double,int*);
extern int  unif_int(int);

#define DEF_VAR(X,Y) \
  SEXP X;PROTECT(X = Y);



int EquiEnergyInit(ee_ctx*ctx,SEXP X) {
  if(X == R_NilValue) return 0;
  ctx->T         = REAL(GET_SLOT(X,install("temperature")))[0];
  ctx->H         = REAL(GET_SLOT(X,install("energy")))[0];
  if(getAttrib(X,install("partition")) != R_NilValue) {
    ctx->partition = REAL(GET_SLOT(X,install("partition")));
    ctx->rings     = GET_SLOT(X,install("rings"));
    ctx->chains    = length(GET_SLOT(X,install("partition")));
    for(ctx->index=0;ctx->index<ctx->chains;ctx->index++) if(ctx->partition[ctx->index] == ctx->H) break;
  }
  ctx->scores    = REAL(GET_SLOT(X,install("score")));
  ctx->population= GET_SLOT(X,install("population"));
  return 1;
}

SEXP BNSL_FastOrderSample(SEXP d,SEXP max_parents,SEXP _N,
			  SEXP order,
			  SEXP ret,SEXP Q,SEXP X,SEXP S,SEXP _usemax) {
  char *vmax         = vmaxget();
  fast_score_ctx ctx = fast_ctx(d,INTEGER(max_parents)[0]); 
  GetRNGstate();
  int B = INTEGER(_N)[0];
  int N = B+INTEGER(_N)[1];
  
  int  i,j;
  int  anneal,anneal_n,n;
  int  accept_jump,reject_jump;
  int *accept_swap,*reject_swap;
  int  usemax = LOGICAL(_usemax)[0];

  double T,anneal_a,score,swap_prob,jump_prob,min_seen;
  SEXP scores,orders;

  swap_prob   = jump_prob   = 0.0;
  anneal_n    = anneal      = 0;
  accept_jump = reject_jump = 0;

  ee_ctx ee,je;
  EquiEnergyInit(&ee,ret);	/* Ourselves */
  if(EquiEnergyInit(&je,X) && B > 0) {
    anneal    = 1;
    anneal_n  = B/2;
    jump_prob = REAL(Q)[0];
    anneal_a  = (log(ee.T)-log(je.T))/((double)anneal_n);
  }
  T = ee.T;

  PROTECT(ret = duplicate(ret));
  PROTECT(scores = allocVector(REALSXP,N));
  SET_SLOT(ret,install("score"),scores);
  PROTECT(orders = allocVector(VECSXP,N));
  SET_SLOT(ret,install("population"),orders);

  n       = ctx->n;
  if(order == R_NilValue || length(order) < n)
    order = OrderRandom(ctx->n);

  min_seen = score   = -fast_ctx_score(ctx,INTEGER(order),n);
  swap_prob = REAL(S)[0];
  accept_swap = (int*)R_alloc(sizeof(int),n+1);
  reject_swap = (int*)R_alloc(sizeof(int),n+1);

  memset(accept_swap,0,sizeof(int)*(n+1));
  memset(reject_swap,0,sizeof(int)*(n+1));

  Rprintf("p(G)  = %.2f*(%.2f/1-%.2f)^|G|\n",ctx->gamma,ctx->beta,ctx->beta);
  Rprintf("pi(x) = exp(-max(h(x),%.2f)/%.2f\n",ee.H,ee.T);
  Rprintf("Swap probability %.2f, jump probability %.2f, anneal %d, burn-in %d, steps %d\n",
	  swap_prob,jump_prob,anneal_n,B,N-B);
  for(i=0;i<N;i++) {
    SEXP proposal          = R_NilValue;
    double proposed_score;
    if(anneal & i <= anneal_n)
      T = je.T*exp(anneal_a*(double)i);

    if(X != R_NilValue && unif_rand() < jump_prob) {
      int ring;SEXP indices;
      for(ring=0;ring<je.chains;ring++)
	if(usemax?Max(score,ee.H):score < je.partition[ring]) break;
      indices = VECTOR_ELT(je.rings,ring);
      if(length(indices) > 0) {
	double lr;
	int    index   = unif_int(length(indices));
	index          = INTEGER(indices)[index];
	proposed_score = je.scores[index];
	lr             = LogTargetRatio(proposed_score,score,ee.H,T)
	  +LogTargetRatio(score,proposed_score,je.H,je.T);
	if(lr >= 0.0 || unif_rand() < exp(lr)) {
	  accept_jump++;
	  PROTECT(proposal = duplicate(VECTOR_ELT(je.population,index)));
	} else {
	  reject_jump++;
	}
      }
    }
    if(proposal == R_NilValue) {
      int c = 0;
      PROTECT(proposal = OrderMetropolisSwapProb(order,swap_prob,&c));
      proposed_score   = -fast_ctx_score(ctx,INTEGER(proposal),n);
      if(proposed_score <= score ||
	 unif_rand()    <= exp(LogTargetRatio(proposed_score,score,ee.H,T))) {
	accept_swap[c]++;
      } else {
	reject_swap[c]++;
	UNPROTECT(1);proposal = R_NilValue;
      }
    }

    if(proposal == R_NilValue) {
      order = duplicate(order);
      SET_VECTOR_ELT(orders,i,order);
    } else {
      SET_VECTOR_ELT(orders,i,proposal);
      UNPROTECT(1);
      score = proposed_score;
      order = proposal;
    }
    REAL(scores)[i] = score;
    if(score < min_seen) min_seen = score;
    if(i % 5000 == 0) {
      Rprintf("%d (%.2f) iterations, %.4f low score, %d (%.2f) jumps\n\t",
	      i,T,min_seen,accept_jump,accept_jump+reject_jump?
	      ((double)accept_jump/((double)accept_jump+(double)reject_jump)):0.0);
      for(j=0;j<n;j++) {
	if(accept_swap[j]+reject_swap[j]>0) 
	  Rprintf("%d:%d (%.2f%) ",j,accept_swap[j]+reject_swap[j],
		  100.0*((double)accept_swap[j]/((double)accept_swap[j]+(double)reject_swap[j])));
      }
      Rprintf("\n");
    }
  }
  {
    Rprintf("%d (%.2f) iterations, %.4f low score, %d (%.2f) jumps\n\t",
	    i,T,min_seen,accept_jump,accept_jump+reject_jump?
	    ((double)accept_jump/((double)accept_jump+(double)reject_jump)):0.0);
    for(j=0;j<n;j++) {
      if(accept_swap[j]+reject_swap[j]>0) 
	Rprintf("%d:%d (%.2f%) ",j,accept_swap[j]+reject_swap[j],
		100.0*((double)accept_swap[j]/((double)accept_swap[j]+(double)reject_swap[j])));
    }
    Rprintf("\n");
  }
  PutRNGstate();
  ee_ChainDiagnostics(ret,accept_jump,reject_jump,accept_swap,reject_swap,ctx->n,INTEGER(_N)[0]);
  vmaxset(vmax);
  UNPROTECT(3);
  return ret;
}



