#include <Rdefines.h>
#include "common.h"
#include "bitstring.h"
#include <db.h>
#include <Rmath.h>
#include "fastdb.h"

extern SEXP OrderRandom(int);
extern SEXP OrderMetropolisProposal(SEXP,int,int*,int*);
extern SEXP OrderMetropolisSwap(SEXP);
extern SEXP OrderMetropolisCut(SEXP);
extern SEXP OrderMetropolisSwapProb(SEXP,double,int*);
extern int  unif_int(int);

SEXP BNSL_FastOrderPT(SEXP d,SEXP max_parents,SEXP _N,SEXP ret,SEXP Q) {
  fast_score_ctx ctx = fast_ctx(d,INTEGER(max_parents)[0]);
  SEXP    temp = GET_SLOT(ret,install("temperature"));
  int     r    = length(temp);
  double *T    = REAL(temp);
  double  q    = REAL(Q)[0];
  int     N    = length(_N) == 2 ? INTEGER(_N)[0]+INTEGER(_N)[1] : INTEGER(_N)[0];
  double *min_seen  = (double*)R_alloc(sizeof(double),r);

  PROTECT(ret = duplicate(ret));
  SEXP pop,scores;
  PROTECT(pop    = allocMatrix(VECSXP,r,N));
  PROTECT(scores = allocMatrix(REALSXP,r,N));
  SET_SLOT(ret,install("population"),pop);
  SET_SLOT(ret,install("score"),scores);
  
  int i,j,ndx = 0;
  int n = ctx->n;
  
  GetRNGstate();

  /* Initialize */
  for(i=0;i<r;i++) {
    SET_VECTOR_ELT(pop,ndx,OrderRandom(n));
    REAL(scores)[ndx] = -fast_ctx_score(ctx,INTEGER(VECTOR_ELT(pop,ndx)),n);
    min_seen[i] = REAL(scores)[ndx];
    ndx++;
  }

  Rprintf("Starting sampler with %d parallel chains (swap p=%.2f)\n",r,q);
  
  /* Run the Sampler */
  for(i=0;i<N;i++) {
    
    if(unif_rand() < q) {
      int a = unif_int(r-1);
      /* Swap Step */
      for(j=0;j<r;j++) {
	SET_VECTOR_ELT(pop,ndx+j,duplicate(VECTOR_ELT(pop,ndx-(r-j))));
	REAL(scores)[ndx+j] = REAL(scores)[ndx-(r-j)];
      }
      if(unif_rand() < exp(-((REAL(scores)[ndx+(a+1)]-REAL(scores)[ndx+a])/T[a]+
			     (REAL(scores)[ndx+a]-REAL(scores)[ndx+(a+1)])/T[a+1]))) {
	SEXP   t1 = VECTOR_ELT(pop,ndx+a);
	double t2 = REAL(scores)[ndx+a];
	SET_VECTOR_ELT(pop,ndx+a,VECTOR_ELT(pop,ndx+a+1));
	SET_VECTOR_ELT(pop,ndx+a+1,t1);
	REAL(scores)[ndx+a] = REAL(scores)[ndx+a+1];
	REAL(scores)[ndx+a+1] = t2;
      }
      ndx += j;
    } else {
      /* Parallel Step */
      SEXP proposal;
      double proposed_score;
      for(j=0;j<r;j++) {
	PROTECT(proposal = OrderMetropolisSwap(VECTOR_ELT(pop,ndx-(r-j))));
	proposed_score   = -fast_ctx_score(ctx,INTEGER(proposal),n);
	if(unif_rand() < exp(-(proposed_score-REAL(scores)[ndx-(r-j)])/T[j])) {
	  SET_VECTOR_ELT(pop,ndx,proposal);
	  REAL(scores)[ndx] = proposed_score;
	} else {
	  SET_VECTOR_ELT(pop,ndx,duplicate(VECTOR_ELT(pop,ndx-(r-j))));
	  REAL(scores)[ndx] = REAL(scores)[ndx-(r-j)];
	}
	if(REAL(scores)[ndx] < min_seen[j]) min_seen[j] = REAL(scores)[ndx];
	UNPROTECT(1);
	ndx++;
      }
      /* End Parallel Step */
    }

    if(i && (i % 5000 == 0)) {
      Rprintf("%d iteration ",i);
      for(j=0;j<r;j++)
	Rprintf("%.2f ",min_seen[j]);
      Rprintf("\n");
    }
  }
  UNPROTECT(3);
  PutRNGstate();
  return ret;
}

