#include "common.h"
#include "bitstring.h"
#include <Rinternals.h>
#include <Rdefines.h>
#include <db.h>
#include "scoring.h"
#include "dag.h"
#include <R_ext/Utils.h>

double dag_distance(int n,char *a,char *b) {
  int d       = bitstr_size(n);
  double dist = 0.0;
  int i,j;
  for(i=0;i<n;i++) {
    for(j=0;j<n;j++) {
      if(bit_test(a+d*i,j) != bit_test(b+d*i,j)) {
	dist += (bit_test(a+d*j,i) != bit_test(a+d*j,i))?0.5:1.0;
      }
    }
  }
  return dist;
}

SEXP BNSL_DagDistMat(SEXP ds) {
  SEXP dags = GET_SLOT(ds,install("dags"));
  int  n    = asInteger(GET_SLOT(ds,install("vertices")));
  int  l    = length(dags);
  SEXP dist;

  PROTECT(dist = allocVector(REALSXP,n*(n-1)/2));
  int i,k=0;
  for(i=0;i<l;i++) {
    char *a = RAW(VECTOR_ELT(dags,i));
    int j;
    for(j=i+1;j<l;j++) {
      char *b = RAW(VECTOR_ELT(dags,j));
      REAL(dist)[k++] = dag_distance(n,a,b);
    }
  }
  setAttrib(dist,install("class"),ScalarString(mkChar("dist")));
  setAttrib(dist,install("Size"),ScalarReal(l));
  setAttrib(dist,install("Diag"),ScalarLogical(FALSE));
  setAttrib(dist,install("Upper"),ScalarLogical(FALSE));
  UNPROTECT(1);
  return dist;
}
