#include <R_ext/RS.h>
#include "common.h"
#include <assert.h>
#include <Rdefines.h>

SEXP BNSL_MDL(SEXP tbl,SEXP child,SEXP set) {
  int i,j,k,l,nc,nr;
  double *XF,*X,*F,Ninv,W,n,ki;
  int    *levels;
  nc     = nlevels(VECTOR_ELT(tbl,INTEGER(child)[0]-1)); /* Table Columns are Child */
  levels = (int*)R_alloc(length(set),sizeof(int));
  nr = 1;
  for(i=0;i<length(set);i++) {
    levels[i] = nlevels(VECTOR_ELT(tbl,INTEGER(set)[i]-1));
    nr       *= levels[i];
  }  

  XF = (double*)R_alloc(sizeof(double),nc*nr);
  F  = (double*)R_alloc(sizeof(double),nr);
  X  = (double*)R_alloc(sizeof(double),nc);
  memset(XF,0,sizeof(double)*nc*nr);
  memset(F,0,sizeof(double)*nr);
  memset(X,0,sizeof(double)*nc);

  l   = length(VECTOR_ELT(tbl,0));
  Ninv = 1.0/(double)l;
  n = (double)length(tbl);
  ki= (double)length(set);
  
  for(i=0;i<l;i++) {
    int col = INTEGER(VECTOR_ELT(tbl,INTEGER(child)[0]-1))[i]-1;
    int row=0;
    int mult=nr;
    for(j=0;j<length(set);j++) {
      int  e = INTEGER(VECTOR_ELT(tbl,INTEGER(set)[j]-1))[i]-1;
      mult /= levels[j];
      row  += mult*e;
    }
    XF[(row*nc)+col] += Ninv;
     F[row] += Ninv;
     X[col] += Ninv;
  }  
  
  k=0;W = 0;
  for(i=0;i<nr;i++) {
    for(j=0;j<nc;j++,k++) {
      W += (XF[k] > 0) ? XF[k]*log(XF[k]/(F[i]))/M_LN2 : 0.0;
    }
  }

  W /= -Ninv;
  W += ((double)nr*(nc-1))*log((double)(l))/M_LN2;
  W += ki*log(n)/M_LN2;

  return ScalarReal(W);



}

SEXP BNSL_Laplace(SEXP tbl,SEXP child,SEXP set) {
  int     i,j,k,l,nc,nr;
  double *mat,result;
  int    *levels;
  nc     = nlevels(VECTOR_ELT(tbl,INTEGER(child)[0]-1)); /* Table Columns are Child */
  levels = (int*)R_alloc(length(set),sizeof(int));
  nr = 1;
  for(i=0;i<length(set);i++) {
    levels[i] = nlevels(VECTOR_ELT(tbl,INTEGER(set)[i]-1));
    nr       *= levels[i];
  }  
  mat = (double*)R_alloc(nc*nr,sizeof(double));  
  memset(mat,0,sizeof(double)*nc*nr);
  l   = length(VECTOR_ELT(tbl,0));
  /*printf("%d ",INTEGER(child)[0]-1);
    for(j=0;j<length(set);j++) printf(" %d",INTEGER(set)[j]-1);
    printf("\n");*/
  
  for(i=0;i<l;i++) {
    int col = INTEGER(VECTOR_ELT(tbl,INTEGER(child)[0]-1))[i]-1;
    int row=0;
    int mult=nr;
    for(j=0;j<length(set);j++) {
      int  e = INTEGER(VECTOR_ELT(tbl,INTEGER(set)[j]-1))[i]-1;
      mult /= levels[j];
      row  += mult*e;
    }
    mat[(row*nc)+col]++;
  }

  result = 0;
  k=0;
  for(i=0;i<nr;i++) {
    double nij=0;
    for(j=0;j<nc;j++,k++) {
      if(mat[k]) result += mat[k]*log(mat[k]);
      nij += mat[k];
    }
    result -= nij*log(nij);
  } 
  return ScalarReal(result);
}

SEXP BNSL_BIC(SEXP tbl,SEXP child,SEXP set) { 
  int i,j,k,l,nc,nr;
  double *XF,*X,*F,Ninv,W,n,ki;
  int    *levels;
  nc     = nlevels(VECTOR_ELT(tbl,INTEGER(child)[0]-1)); /* Table Columns are Child */
  levels = (int*)R_alloc(length(set),sizeof(int));
  nr = 1;
  for(i=0;i<length(set);i++) {
    levels[i] = nlevels(VECTOR_ELT(tbl,INTEGER(set)[i]-1));
    nr       *= levels[i];
  }  

  XF = (double*)R_alloc(sizeof(double),nc*nr);
  F  = (double*)R_alloc(sizeof(double),nr);
  X  = (double*)R_alloc(sizeof(double),nc);
  memset(XF,0,sizeof(double)*nc*nr);
  memset(F,0,sizeof(double)*nr);
  memset(X,0,sizeof(double)*nc);

  l   = length(VECTOR_ELT(tbl,0));
  Ninv = 1.0/(double)l;
  n = (double)length(tbl);
  ki= (double)length(set);
  
  for(i=0;i<l;i++) {
    int col = INTEGER(VECTOR_ELT(tbl,INTEGER(child)[0]-1))[i]-1;
    int row=0;
    int mult=nr;
    for(j=0;j<length(set);j++) {
      int  e = INTEGER(VECTOR_ELT(tbl,INTEGER(set)[j]-1))[i]-1;
      mult /= levels[j];
      row  += mult*e;
    }
    XF[(row*nc)+col] += Ninv;
     F[row] += Ninv;
     X[col] += Ninv;
  }  
  
  k=0;W = 0;
  for(i=0;i<nr;i++) {
    for(j=0;j<nc;j++,k++) {
      W += (XF[k] > 0) ? XF[k]*log(XF[k]/(F[i]))/M_LN2 : 0.0;
    }
  }

  W /= -Ninv;
  W += ((double)nr*(nc-1))*log((double)(l))/M_LN2;
  W += ki*log(n)/M_LN2;

  return ScalarReal(-W);
}

/**
 * An efficient (but not terribly readable) implementation of the 
 * K2 algorithm. It works with no modification of the original data
 * or allocation of extra memory aside from the SEXP for the result.
 * Used with DUP=FALSE it should be fast and efficient
 */
SEXP BNSL_K2(SEXP tbl,SEXP child,SEXP set,SEXP alpha) {
  
  int     i,j,k,l,nc,nr;
  double *mat,result,beta;
  int    *levels;
  beta   = (1-REAL(alpha)[0])*(1-REAL(alpha)[0]);
  nc     = nlevels(VECTOR_ELT(tbl,INTEGER(child)[0]-1)); /* Table Columns are Child */
  levels = (int*)R_alloc(length(set),sizeof(int));
  nr = 1;
  for(i=0;i<length(set);i++) {
    levels[i] = nlevels(VECTOR_ELT(tbl,INTEGER(set)[i]-1));
    nr       *= levels[i];
  }
  
  mat = (double*)R_alloc(nc*nr,sizeof(double));
  memset(mat,0,sizeof(double)*nc*nr);
  l   = length(VECTOR_ELT(tbl,0));
  /*printf("%d ",INTEGER(child)[0]-1);
    for(j=0;j<length(set);j++) printf(" %d",INTEGER(set)[j]-1);
    printf("\n");*/
  
  for(i=0;i<l;i++) {
    int col = INTEGER(VECTOR_ELT(tbl,INTEGER(child)[0]-1))[i]-1;
    int row=0;
    int mult=nr;
    for(j=0;j<length(set);j++) {
      int  e = INTEGER(VECTOR_ELT(tbl,INTEGER(set)[j]-1))[i]-1;
      mult /= levels[j];
      row  += mult*e;
    }
    mat[(row*nc)+col]++;
  }
  
  k=0;
  result = lgamma((double)nc);
  for(i=0;i<nr;i++) {
    double nij=0;
    for(j=0;j<nc;j++,k++) {
      result += lgamma(beta*mat[k]+1.0);
      nij += mat[k];
      /*      Rprintf("n_ijk=%.0f\n",mat[k]);*/
    }
    /*    Rprintf("n_ij=%.0f\n",nij);*/
    result -= lgamma(beta*nij+(double)nc);
  }
  return ScalarReal(result);
}

void build_mat(SEXP tbl,SEXP _child,SEXP set,SEXP _skip,SEXP _lag,
	       double **matrix,int*numRows,int*numCols) {
  int  skipOn = length(_skip);
  int *skip   = LOGICAL(AS_LOGICAL(_skip));
  int  lag    = asInteger(_lag);

  //Construct the child list
  SEXP child  = VECTOR_ELT(tbl,INTEGER(_child)[0]-1);
  int  nc;
  nc = nlevels(child);
  if(nc == 0) {
    if((nc = length(getAttrib(child,install("levels")))) == 0)
      error("Unable to obtain a count of levels for the child column");
  }

  //Construct the parent set 
  SEXP pset;
  int *levels;
  int  np,nr;
  np = length(set);
  levels = (int*)R_alloc(sizeof(int),np);
  nr     = 1;
  PROTECT(pset = allocVector(VECSXP,np));
  int i;
  for(i=0;i<np;i++) {
    SET_VECTOR_ELT(pset,i,AS_INTEGER(VECTOR_ELT(tbl,INTEGER(set)[i]-1)));
    int l = nlevels(VECTOR_ELT(pset,i));
    if(l == 0) {
      if((l = length(getAttrib(VECTOR_ELT(pset,i),install("levels")))) == 0)
	error("Unable to obtain levels from parent (%d)",INTEGER(set)[i]);
    }
    levels[i] = l;
    nr       *= l;
  }
  
  double *mat = (double*)R_alloc(sizeof(double),nr*nc);
  memset(mat,0,sizeof(double)*nr*nc);
  int l = length(child);
  for(i=lag;i<l;i++) {
    int col  = INTEGER(child)[i]-1;
    int row  = 0;
    int mult = nr;
    int j;
    for(j=0;j<np;j++) {
      int  e = INTEGER(VECTOR_ELT(pset,j))[i-lag]-1;
      mult /= levels[j];
      row  += mult*e;
    }
    if(skipOn == 0 || (skipOn == l && skip[i] != FALSE)) 
      mat[(row*nc)+col] += 1.0;
  }
  // Return results
  matrix[0]  = mat;
  numRows[0] = nr;
  numCols[0] = nc;
  UNPROTECT(1);
}

SEXP BNSL_BDeu(SEXP tbl,SEXP child,SEXP set,SEXP _alpha,
	       SEXP _skip,SEXP _lag) {
  int     i,j,k,nr,nc;
  double *mat;
  char   *vmax = vmaxget();
  build_mat(tbl,child,set,_skip,_lag,&mat,&nr,&nc);
  double alpha = asReal(_alpha);
  double result;

  double aij = alpha/((double)nr*nc);
  double ai  = alpha/((double)nr);
  double qi  = (double)nr;
  double qiri= (double)nr*nc;
  k=0;
  result = lgamma(ai)*qi - lgamma(aij)*qiri;
  for(i=0;i<nr;i++) {
    double ni = 0;
    for(j=0;j<nc;j++,k++) {
      result += lgamma(aij+mat[k]);
      ni     += mat[k];
    }
    result -= lgamma(ai+ni);
  }
  vmaxset(vmax);
  return ScalarReal(result);
}


SEXP BNSL_BDeu_old(SEXP tbl,SEXP child,SEXP set,SEXP sp,SEXP alpha,
	       SEXP skip,SEXP _lag) {
  
  int     i,j,k,l,nc,nr;
  double *mat,result,beta;
  int         *levels;
  int    *skipme = LOGICAL(skip);
  int     skipOn = length(skip);
  int     lag    = asInteger(_lag);
  beta   = (1-REAL(alpha)[0])*(1-REAL(alpha)[0]);
  nc     = nlevels(VECTOR_ELT(tbl,INTEGER(child)[0]-1)); /* Table Columns are Child */
  levels = (int*)R_alloc(length(set),sizeof(int));
  nr = 1;
  for(i=0;i<length(set);i++) {
    levels[i] = nlevels(VECTOR_ELT(tbl,INTEGER(set)[i]-1));
    nr       *= levels[i];
  }
  
  mat = (double*)R_alloc(nc*nr,sizeof(double));
  memset(mat,0,sizeof(double)*nc*nr);
  l   = length(VECTOR_ELT(tbl,0));
  /*printf("%d ",INTEGER(child)[0]-1);
    for(j=0;j<length(set);j++) printf(" %d",INTEGER(set)[j]-1);
    printf("\n");*/
  
  for(i=lag;i<l;i++) {
    int col = INTEGER(VECTOR_ELT(tbl,INTEGER(child)[0]-1))[i]-1;
    int row=0;
    int mult=nr;
    for(j=0;j<length(set);j++) {
      int  e = INTEGER(VECTOR_ELT(tbl,INTEGER(set)[j]-1))[i-lag]-1;
      mult /= levels[j];
      row  += mult*e;
    }
    if(skipOn == 0 || (skipOn == l && skipme[i] != FALSE)) mat[(row*nc)+col]++;
  }
  
  k=0;
  result = log(REAL(sp)[0])*((double)(nc-1)*nr)
    + ((double)nr)*lgamma(REAL(alpha)[0]/(double)nr)
    - ((double)nr*nc)*lgamma(REAL(alpha)[0]/((double)nr*nc));
  
  for(i=0;i<nr;i++) {
    double nij=0;
    for(j=0;j<nc;j++,k++) {
      result += lgamma(REAL(alpha)[0]/((double)nr*nc)+mat[k]);
      nij += mat[k];
      /*      Rprintf("n_ijk=%.0f\n",mat[k]);*/
    }
    /*    Rprintf("n_ij=%.0f\n",nij);*/
    result -= lgamma(nij+REAL(alpha)[0]/((double)nr));
  }
  return ScalarReal(result);
}
