#include "common.h"
#include <Rdefines.h>
#include <db.h>
#include "scoring.h"
#include "eesample.h"


#define Max(X,Y) (((X)>(Y))?(X):(Y))

#define LogTargetRatio(Hnew,Hcur,Hi,T) \
  (Max((Hcur),(Hi))-Max((Hnew),(Hi)))/(T)

extern SEXP OrderRandom(int);
extern SEXP OrderMetropolisProposal(SEXP,int,int*,int*);
extern SEXP OrderMetropolisSwap(SEXP);
extern SEXP OrderMetropolisCut(SEXP);
extern SEXP OrderMetropolisSwapProb(SEXP,double,int*);
extern int  unif_int(int);

/* 
 * @param N  Integer, N[0] = burn-in,N[1] = samples to keep
 */
SEXP BNSL_OrderEquiEnergySample(SEXP d,SEXP max_parents,SEXP mode,
				SEXP _N,SEXP order,
				SEXP xH,SEXP xT,
				SEXP Q,SEXP X,SEXP S,SEXP unique,SEXP tladder) {
  char*     vmax = vmaxget();
  score_ctx ctx = bnsl_score_ctx(d,INTEGER(max_parents)[0],INTEGER(mode)[0]);
  int       accept_jump,reject_jump;
  int      *accept_swap,*reject_swap;
  int       i,j,N,B;
  double  H,T,swap_prob;
  double  min_seen;  
  double    score;
  SEXP      res,scores,orders;


  /* From the incoming landscape */
  int     chain_n;
  SEXP    chain_rings;
  double* chain_part;
  double* chain_scores;
  SEXP    chain_pop;
  double  chain_temp,chain_energy;



  /* Annealing Schedule */
  char anneal     = 0;
  double anneal_a;
  int    anneal_n = 0;
  accept_jump = reject_jump = 0;
  B           = INTEGER(_N)[0];
  N           = INTEGER(_N)[0]+INTEGER(_N)[1];
  H           = REAL(xH)[0];
  T           = REAL(xT)[0];
  swap_prob    = REAL(S)[0];
  min_seen    = R_PosInf;

  if(X != R_NilValue) {
    Rprintf("Extracting info from X\n");
    chain_temp   = REAL(GET_SLOT(X,install("temperature")))[0];
    chain_energy = REAL(GET_SLOT(X,install("energy")))[0];

    chain_n      = length(GET_SLOT(X,install("partition")));
    chain_part   = REAL(GET_SLOT(X,install("partition")));
    chain_rings  = GET_SLOT(X,install("rings"));

    chain_scores = REAL(GET_SLOT(X,install("score")));
    chain_pop    = GET_SLOT(X,install("population"));

    if(B > 0) {
      anneal = 1;
      anneal_n = B/2;
      anneal_a = (log(T)-log(chain_temp))/((double)anneal_n);
    }

  }


  PROTECT(res    = NEW_OBJECT(MAKE_CLASS("sample")));
  PROTECT(scores = allocVector(REALSXP,N));
  SET_SLOT(res,install("score"),scores);
  PROTECT(orders = allocVector(VECSXP,N));
  /* Assign the population to be of orders */
  classgets(orders,ScalarString(mkString("order.population")));
  SET_SLOT(res,install("population"),orders);

  /* Two diagnostic bits */
  SET_SLOT(res,install("temperature"),ScalarReal(T));
  SET_SLOT(res,install("energy"),ScalarReal(H));
  GetRNGstate();  
  /* Initialize the chain */
  if(order == R_NilValue || length(order) == 0)
    order = OrderRandom(ctx->n);
  score = -bnsl_score_ordering(ctx,INTEGER(order),ctx->n);
  Rprintf("pi(x) = exp(-max(h(x),%.2f)/%.2f)\n",H,T);
  Rprintf("Swap probability %.2f, anneal %d, burn-in %d, steps %d\n",
	  swap_prob,anneal_n,B,N-B);
  Rprintf("Starting Score: %.2f (%.2f %.2f)\n",score,ctx->prior.beta,ctx->prior.gamma);

  accept_swap = (int*)R_alloc(sizeof(int),ctx->n);
  reject_swap = (int*)R_alloc(sizeof(int),ctx->n);

  memset(accept_swap,0,sizeof(int)*ctx->n);
  memset(reject_swap,0,sizeof(int)*ctx->n);

  /* Run the chain */
    for(i=0;i<N;i++) {
    SEXP proposal        = R_NilValue;
    double proposed_score;

    if(anneal & i <= anneal_n)
      T = chain_temp*exp(anneal_a*(double)i);

    if(X != R_NilValue && unif_rand() < REAL(Q)[0]) {
      int ring;SEXP indices;
      for(ring=0;ring<chain_n;ring++)
	if(score < chain_part[ring]) break;
      indices = VECTOR_ELT(chain_rings,ring);
      if(length(indices)>20) {
	double lr;
	int    index=unif_int(length(indices));
	index       =INTEGER(indices)[index];

	proposed_score = chain_scores[index];
	lr             = LogTargetRatio(proposed_score,score,H,T)
	  +LogTargetRatio(score,proposed_score,chain_energy,chain_temp);
	if(lr >= 0.0 || unif_rand() < exp(lr)) {
	  accept_jump++;
	  PROTECT(proposal = duplicate(VECTOR_ELT(chain_pop,index)));
	} else {
	  reject_jump++;
	}
      }
    }
    
    if(proposal == R_NilValue) {
      int c=0;
      PROTECT(proposal = OrderMetropolisSwapProb(order,swap_prob,&c));
      proposed_score = -bnsl_score_ordering(ctx,INTEGER(proposal),ctx->n);
      if(proposed_score <= score ||
	 unif_rand()    <= exp(LogTargetRatio(proposed_score,score,H,T))) {
	accept_swap[c]++;
      } else {
	reject_swap[c]++;
	UNPROTECT(1);proposal = R_NilValue;
      }
    }
    
    if(proposal == R_NilValue) {
      order = duplicate(order);
      SET_VECTOR_ELT(orders,i,order);
    } else {
      SET_VECTOR_ELT(orders,i,proposal);
      UNPROTECT(1);
      score = proposed_score;
      order = proposal;
    }
    REAL(scores)[i] = score;
    if(score < min_seen) min_seen =score;
    
    if(i % 1000 == 0 && i) {
      Rprintf("%d (%.2f) iterations, %.4f low score, %d (%.2f) jumps\n\t",
	      i,T,min_seen,accept_jump,(accept_jump+reject_jump)?
	      ((double)accept_jump)/((double)accept_jump+(double)reject_jump):0.0);
      for(j=0;j<ctx->n;j++) {
	Rprintf("%d/%d ",accept_swap[j],reject_swap[j]);
      }
      Rprintf("\n");
    }

  }
  Rprintf("%d (%.2f) iterations, %.4f low score, %d (%.2f) jumps\n\t",
	  i,T,min_seen,accept_jump,(accept_jump+reject_jump)?
	  ((double)accept_jump)/((double)accept_jump+(double)reject_jump):0.0);
  for(j=0;j<ctx->n;j++) {
    Rprintf("%d/%d ",accept_swap[j],reject_swap[j]);
  }
  Rprintf("\n");
  PutRNGstate();
  ee_ChainDiagnostics(res,accept_jump,reject_jump,
		      accept_swap,reject_swap,ctx->n,
		      INTEGER(_N)[0]);
  bnsl_score_ctx_done(ctx);
  vmaxset(vmax);
  UNPROTECT(3);
  return res;
}
