#include "common.h"
#include <Rdefines.h>
#include <db.h>
#include "scoring.h"
extern int unif_int(int);

SEXP OrderEquiEnergyJump(double original,SEXP sample) {
  int  ring;
  SEXP ringNdx;
  
  SEXP maxEnergy = VECTOR_ELT(sample,0); /* Maximum energy parition */

  /* Find the appropriate energy ring */
  for(ring=0;ring<length(maxEnergy);ring++)
    if(original < REAL(maxEnergy)[ring]) break;
  /* Get the ring membership indices */
  ringNdx = VECTOR_ELT(VECTOR_ELT(sample,1),ring);
  if(ringNdx == R_NilValue || length(ringNdx) == 0)
    return R_NilValue;
  /* Return a randomly selected sample from the ring */
  return duplicate(VECTOR_ELT(VECTOR_ELT(sample,2),INTEGER(ringNdx)[unif_int(length(ringNdx))]-1));
}

SEXP OrderRandom(int n) {
  int  i;
  SEXP order;
  PROTECT(order = allocVector(INTSXP,n));
  for(i=0;i<n;i++) INTEGER(order)[i] = i+1;
  /* Fisher-Yates Shuffle */
  for(i=0;i<n;i++) {
    int j = i+unif_int(n-i);
    int t = INTEGER(order)[j];
    INTEGER(order)[j] = INTEGER(order)[i];
    INTEGER(order)[i] = t;
  }
  UNPROTECT(1);
  return order;
}

SEXP OrderMetropolisSwap(SEXP original) {
  SEXP proposed;
  int  a,b,t,n=length(original);
  PROTECT(proposed = duplicate(original));
  a = unif_int(n);b = a+unif_int(n-a);
  t = INTEGER(proposed)[a];
  INTEGER(proposed)[a] = INTEGER(proposed)[b];
  INTEGER(proposed)[b] = t;
  UNPROTECT(1);
  return proposed;
}

SEXP OrderMetropolisSwapProb(SEXP original,double p,int*swap_c) {
  char *vmax  = vmaxget();
  int   i,k,n     = length(original);
  int  *swaps = (int*)R_alloc(sizeof(int),n);
  SEXP  proposed;
  PROTECT(proposed = duplicate(original));
  k=0;
  if(p>0) {
    for(i=0;i<n;i++) 
      if(unif_rand()<p) swaps[k++] = i;
  }
  if(k < 2) {
    i = unif_int(n);k = i+unif_int(n-i);
    INTEGER(proposed)[i] = INTEGER(original)[k];
    INTEGER(proposed)[k] = INTEGER(original)[i];
    if(swap_c) *swap_c=2;
  } else {
    /* Rotate rather than swap so that odd valued swaps might be more helpful */
    for(i=0;i<k;i++)
      INTEGER(proposed)[swaps[i]] = INTEGER(original)[swaps[(i+1)%k]];
    if(swap_c) *swap_c=k;
  }
  vmaxset(vmax);
  UNPROTECT(1);
  return proposed;
}

SEXP OrderMetropolisCut(SEXP original) {
  SEXP proposed;
  int  n=length(original),a=unif_int(n);
  PROTECT(proposed = allocVector(INTSXP,n));
  memcpy(INTEGER(proposed),&(INTEGER(original)[a]),sizeof(int)*(n-a));
  memcpy(&(INTEGER(proposed)[n-a]),INTEGER(original),sizeof(int)*a);
  UNPROTECT(1);
  return proposed;
}

SEXP OrderMetropolisProposal(SEXP original,int swaps,int*first,int*last) {
  SEXP proposed;
  int  i,c,a,n,t;
  char *vmax;
  int  *mark;
  int  *sel;
  PROTECT(proposed = duplicate(original));
  n = length(proposed);
  c = unif_int(swaps);if(c<1) c = 2; else c *= 2;
  vmax = (char*)vmaxget();
  mark = (int*)R_alloc(sizeof(int),n);
  sel  = (int*)R_alloc(sizeof(int),c);
  *first = n;*last = 0;
  for(i=0;i<n;i++) mark[i] = i;
  for(i=0;i<c;i++) { 
    a=i+unif_int(n-i);t=mark[a];mark[a]=mark[i];mark[i]=t;
    if(mark[i] < *first) *first = mark[i];
    if(mark[i] > *last)  *last  = mark[i];
  }
  for(i=0;i<c;i++) INTEGER(proposed)[mark[i]] = INTEGER(original)[mark[(c-i)-1]];
/*
  memset(mark,0,sizeof(int)*n);
  for(i=0;i<c;i++) {
    do { a = unif_int(n); } while(mark[a]==1);
    mark[a] = 1;sel[i] = a;
  }
  for(i=0;i<c;i++)
    INTEGER(proposed)[sel[i]] = INTEGER(original)[sel[(c-i)-1]];
*/
  vmaxset(vmax);
  UNPROTECT(1);
  return proposed;
}
