#include <Rdefines.h>
#include "common.h"
#include "bitstring.h"
#include <db.h>
#include "psetdag.h"
#include "eesample.h"
#include <Rmath.h>

#define SWAP(A,B) \
  char t = A; \
  A = B; \
  B = t;

#define Max(X,Y) (((X)>(Y))?(X):(Y))

#define LogTargetRatio(Hnew,Hcur,Hi,T) \
  (Max((Hcur),(Hi))-Max((Hnew),(Hi)))/(T)


extern SEXP Rf_findVar1(SEXP,SEXP,SEXPTYPE,int);
extern int  unif_int(int);

void convert_fastdb(DB*db,DB*newdb,int n,int k) {
  DBT   key,data,key2;
  DBC   *cur;
  int   ret,N = bitstr_size(n);
  char *pset  = (char*)R_alloc(sizeof(char),N);
  
  memset(&key,0,sizeof(DBT));
  memset(&key2,0,sizeof(DBT));
  memset(&data,0,sizeof(DBT));

  key2.data = pset;
  key2.size = sizeof(char)*N;

  db->cursor(db,NULL,&cur,0);
  while(cur->c_get(cur,&key,&data,DB_NEXT) == 0) {
    memset(pset,0,sizeof(char)*N);
    int  i,K = key.size/sizeof(int);
    int   *p = (int*)key.data;
    if(K == 0 || K > k) continue;
    if(p[0] > 0) 
      for(i=0;i<K;i++) bit_set(pset,p[i]-1);
    newdb->put(newdb,NULL,&key2,&data,0);
  }
}

double* pset_get_pset(pset_dag_ctx pd,char *pset) {
  DBT key,data;
  memset(&key,0,sizeof(DBT));
  memset(&data,0,sizeof(DBT));
  key.data = pset;
  key.size = pd->rl;
  if(pd->db->get(pd->db,NULL,&key,&data,0) == 0)
    return data.data;
  else
    return NULL;
}

void pset_dag_clear(pset_dag_ctx pd) {
  pd->edges = 0;
  memset(pd->adjM,0,sizeof(char)*pd->rl*pd->n);
  memset(pd->parent,0,sizeof(char)*pd->n*pd->n);
  memset(pd->child,0,sizeof(char)*pd->n*pd->n);
  memset(pd->fanIn,0,sizeof(char)*pd->n);
  memset(pd->fanOut,0,sizeof(char)*pd->n);
  memset(pd->score,0,sizeof(double)*pd->n);
  if(pd->scoring) 
    memcpy(pd->score,pset_get_pset(pd,pd->adjM),sizeof(double)*pd->n);
}


pset_dag_ctx pset_dag(SEXP D,int k,int kk,int scoring) {
  pset_dag_ctx pd = (pset_dag_ctx)R_alloc(sizeof(struct pset_dag_ctx),1);

  pd->n = INTEGER(getAttrib(D,install("vars")))[0];
  pd->k = k;
  pd->kk= (kk==0)?pd->n:kk;
  pd->scoring = scoring;
  pd->edges   = 0;
  
  pd->rl     = bitstr_size(pd->n);
  pd->adjM   = (char*)R_alloc(sizeof(char),pd->rl*pd->n);
  pd->parent = (char*)R_alloc(sizeof(char),pd->n*pd->n);
  pd->child  = (char*)R_alloc(sizeof(char),pd->n*pd->n);
  pd->fanIn  = (char*)R_alloc(sizeof(char),pd->n);
  pd->fanOut = (char*)R_alloc(sizeof(char),pd->n);
  pd->score  = (double*)R_alloc(sizeof(double),pd->n);
  pd->stack  = (char*)R_alloc(sizeof(char),pd->n*pd->n);
  pd->color  = (char*)R_alloc(sizeof(char),pd->n);

  SEXP check;
  check = Rf_findVar1(install(".Gamma"),R_GlobalEnv,REALSXP,FALSE);
  pd->gamma     = (check == R_UnboundValue || length(check)==0)?0.0:REAL(check)[0];


  //Convert FastDB into a more appropriate form
  if(pd->scoring) {
    db_create(&(pd->db),NULL,0);
    pd->db->open(pd->db,NULL,NULL,NULL,DB_BTREE,DB_CREATE,0);
    convert_fastdb(R_ExternalPtrAddr(VECTOR_ELT(D,0)),pd->db,pd->n,pd->k);

  }
  pset_dag_clear(pd);
  return pd;
}

void pset_dag_dump(pset_dag_ctx ctx) {
  int i;
  int j;
  char *pset = ctx->adjM;
  char *q    = ctx->parent;
  for(i=0;i<ctx->n;i++) {
    for(j=0;j<ctx->fanIn[i];j++) 
      Rprintf("%d%s",q[j],bit_test(pset,q[j])?" ":"* ");
    Rprintf("-> %d\n",i);
    pset += ctx->rl;
    q    += ctx->n;
  }
}

SEXP pset_dag_state(pset_dag_ctx ctx) {
  SEXP ret;
  PROTECT(ret = allocVector(RAWSXP,ctx->rl*ctx->n));
  memcpy(RAW(ret),ctx->adjM,ctx->rl*ctx->n);
}

int pset_dag_cycle(pset_dag_ctx ctx,char a,char b) {
  int sp = 0,i = 0,j;
  memset(ctx->color,0,sizeof(char)*ctx->n);
  ctx->color[a]    = 1;
  ctx->stack[sp++] = a;
  while(i < sp) {
    int   n = ctx->stack[i++];
    char *q = ctx->parent + ctx->n*n;
    for(j=0;j<ctx->fanIn[n];j++) 
      if(q[j] == b) 
	return 1;  
      else if(ctx->color[q[j]] == 0) {
	ctx->stack[sp++] = q[j];
	ctx->color[q[j]] = 1;
      }
  }
  return 0;
}

int pset_dag_edge(pset_dag_ctx ctx,char a,char b) {
  char *pset = ctx->adjM + ctx->rl*b;
  return bit_test(pset,a) ? 1 : 0;
}

double pset_dag_add(pset_dag_ctx ctx,char a,char b,double *old) {
  if(old && ctx->scoring) *old = ctx->score[b];
  char *pset = ctx->adjM + ctx->rl*b;
  //Can't add an edge if it exists or if it introduces a cycle
  if(bit_test(pset,a) ||
     pset_dag_cycle(ctx,a,b) ||
     ctx->fanIn[b]+1 >= ctx->k) return R_NegInf;
  char *q = ctx->parent + ctx->n*b;
  q[ctx->fanIn[b]++] = a;
  q       = ctx->child  + ctx->n*a;
  q[ctx->fanOut[a]++] = b;
  bit_set(pset,a);
  ctx->edges++;
  if(ctx->scoring) {
    double *scores = pset_get_pset(ctx,pset);
    ctx->score[b] = scores[b];
    return ctx->score[b];
  } else
    return R_PosInf;
}

double pset_dag_del(pset_dag_ctx ctx,char a,char b,double *old) {
  int i;
  if(old && ctx->scoring) *old = ctx->score[b];
  char *pset = ctx->adjM + ctx->rl*b;
  if(!bit_test(pset,a)) return R_NegInf;
  char *q = ctx->parent + ctx->n*b;
  for(i=0;i<ctx->fanIn[b];i++)
    if(q[i] == a) break;
  q[i] = q[--ctx->fanIn[b]];
  q = ctx->child + ctx->n*a;
  for(i=0;i<ctx->fanOut[a];i++)
    if(q[i] == b) break;
  q[i] = q[--ctx->fanOut[a]];
  bit_clear(pset,a);
  ctx->edges--;
  if(ctx->scoring) {
    double *scores = pset_get_pset(ctx,pset);
    ctx->score[b] = scores[b];
    return ctx->score[b];
  }
}

double pset_dag_set(pset_dag_ctx ctx,char *adjM) {
  int i,j;
  memcpy(ctx->adjM,adjM,ctx->rl*ctx->n);
  memset(ctx->fanIn,0,sizeof(char)*ctx->n);
  memset(ctx->fanOut,0,sizeof(char)*ctx->n);
  char *pset = ctx->adjM;
  char *par  = ctx->parent;
  ctx->edges = 0;
  for(i=0;i<ctx->n;i++) {
    for(j=0;j<ctx->n;j++) 
      if(bit_test(pset,j)) {
	char *cld = ctx->child + ctx->n*j;
	par[ctx->fanIn[i]++]  = j;
	cld[ctx->fanOut[j]++] = i;
	ctx->edges++;
      }
    if(ctx->scoring) {
      double *scores = pset_get_pset(ctx,pset);
      ctx->score[i] = scores[i];
    }
    par  += ctx->n;
    pset += ctx->rl;
  }
  if(!ctx->scoring) return R_NegInf;
  double score = 0.0;
  for(i=0;i<ctx->n;i++) score += ctx->score[i];
  return score;
}

/* Generate a random parent set */
double pset_dag_random(pset_dag_ctx ctx,char *parents,int n,char b) {
  int i,k,count = 0;
  if(n==0) return ctx->score[b];
/*   for(i=0;i<=ctx->k;i++) { */
/*     ctx->color[i] = choose(n,i); */
/*     count += ctx->color[i]; */
/*   } */
/*   double cdf = 0.0; */
/*   double u   = unif_rand(); */
/*   for(k=0;k<=ctx->k;k++) { */
/*     cdf += ((double)ctx->color[k])/((double)count); */
/*     if(u < cdf) break; */
/*   } */
  k = (ctx->k > n) ? n : k;
  k = unif_int(k+1);
  if(k==0) return ctx->score[b];
  
  /* Select all */
  if(n && k==n) {
    for(i=0;i<n;i++) pset_dag_add(ctx,parents[i],b,NULL);
    return ctx->score[b];
  }
  /* Select a random subset */
  for(i=0;i<n;i++)
    ctx->stack[i] = parents[i];
  for(i=0;i<k;i++) {
    int a = i+unif_int(n-i);
    int t = ctx->stack[a];
    ctx->stack[a] = ctx->stack[i];
    ctx->stack[i] = t;
    pset_dag_add(ctx,t,b,NULL);
  }
  return ctx->score[b];
}





//Generate a more-or-less uniformly random linear extension with
//an approximation of the number of linear extensions
int pset_linear_extension(pset_dag_ctx ctx,char *ext,double *vol,int cont) {
  int i,j,sp = 0,ep = 0;

  if(!cont) {
    //Try not to always get the same starting point for a DAG
    for(i=0;i<ctx->n;i++)
      ext[i] = i;
    for(i=0;i<ctx->n;i++) {
      int a  = i+unif_int(ctx->n-i);
      SWAP(ext[i],ext[a]);
    }
    
    //Do a topological sort
    memcpy(ctx->color,ctx->fanIn,sizeof(char)*ctx->n);
    for(i=0;i<ctx->n;i++)
      if(ctx->fanIn[ext[i]] == 0) ctx->stack[sp++] = ext[i];
    //We can caculate the number of zeros exactly, so do it.
    *vol = gamma(sp+1.0);
    while(sp > 0) {
      char  v   = ctx->stack[--sp];
      char *c   = ctx->child + v*ctx->n;
      ext[ep++] = v;
      for(j=0;j<ctx->fanOut[v];j++) 
	if(ctx->color[c[j]] && --ctx->color[c[j]] == 0) 
	  ctx->stack[sp++] = c[j];
    }
    //Whoa, not a DAG?!
    if(ep < ctx->n) return 0;
  }


  for(i=0;i<ep;i++) 
    if(ctx->fanIn[ext[i]] > 0) break;
  //Now do the approximate calculations
  for(;i<ep;i++) {
    int cur = ext[i];
    int N   = 17*pow(i+1,3)*log(i+1);
    double in = 0.0;
    //We only need to do this if we aren't in a chain setting
    if(!bit_test(ctx->adjM + ctx->rl*cur,ext[i-1])) {
      /*      Rprintf("Before: ");
      char  *pset = ctx->adjM + ctx->rl*cur;
      for(j=0;j<=i;j++)
	Rprintf("%d%c ",ext[j],bit_test(pset,ext[j])?'*':' ');
	Rprintf("\n");*/
      for(j=0;(j<2*N || in == 0.0);j++) {
	if(unif_rand() > 0.5) {
	  int a = unif_int(i);
	  if(!bit_test(ctx->adjM + ctx->rl*ext[a+1],ext[a])) {
	    SWAP(ext[a],ext[a+1]);
	  }
	}
	if(j>N && ext[i] == cur) in += 1.0;
      }
      /*      Rprintf("\nAfter: ");
      for(j=0;j<=i;j++)
      Rprintf("%d ",ext[j]);*/
      *vol *= ((double)N-1.0)/in;
      /*      Rprintf(" (%.2f %.2f) \n",in/((double)N-1.0),*vol); */
    }
  }
  return 1;
}

SEXP pset_make_dag(pset_dag_ctx ctx) {
  SEXP ret;
  PROTECT(ret = allocVector(RAWSXP,ctx->rl*ctx->n));
  memcpy(RAW(ret),ctx->adjM,ctx->rl*ctx->n);
  UNPROTECT(1);
  return ret;
}


typedef struct move {
  char   t,a,b;
  double score;
} move;

double pset_make_move(pset_dag_ctx ctx,move* m) {
  switch(m->t) {
  case 1:return pset_dag_add(ctx,m->a,m->b,NULL);
  case 2:return pset_dag_del(ctx,m->a,m->b,NULL);
  case 3:pset_dag_del(ctx,m->a,m->b,NULL);return pset_dag_add(ctx,m->b,m->a,NULL);
  default:return R_NegInf;
  }
}




int calc_moves(pset_dag_ctx ctx,move* moves,int *ss,double orig,SEXP cand) {
  int _j,i,j,count = 0;
  char *pset = ctx->adjM;
  double score,old;
  for(i=0;i<ctx->n;i++) {
    SEXP r = cand == R_NilValue?R_NilValue:VECTOR_ELT(cand,i);
    for(_j=0;_j<(r==R_NilValue)?ctx->n:length(r);_j++) {
      j = (r==R_NilValue)?_j:INTEGER(cand)[j];
      if(i==j) continue;
      if(bit_test(pset,j)) {
	//Try to remove
	if((score = pset_dag_del(ctx,j,i,&old)) != R_NegInf) {
	  moves[count].t = 2;moves[count].a = j;moves[count].b = i;
	  moves[count++].score = (orig-old + score);
	  if(ss) ss[count-1] = count-1;
	  /* 	  Rprintf("[%d] %d->%d => %d  %d %.2f\n",count-1,j,i,j,i,moves[count-1].score); */
	  //Can we reverse?
	  if((score = pset_dag_add(ctx,i,j,&old)) != R_NegInf) {
	    moves[count].t = 3;moves[count].a = j;moves[count].b = i;
	    moves[count++].score = moves[count-1].score - old + score;
	    if(ss) ss[count-1] = count-1;
	    /* 	    Rprintf("[%d] %d->%d => %d<-%d %.2f\n",count-1,j,i,j,i,moves[count-1].score); */
	    //Put things back
	    pset_dag_del(ctx,i,j,NULL);
	  }
	  //Put things back
	  pset_dag_add(ctx,j,i,NULL);
	}
      } else {
	//We can add
	if((score = pset_dag_add(ctx,j,i,&old)) != R_NegInf) {

	  moves[count].t = 1;moves[count].a = j;moves[count].b = i;
	  moves[count++].score = (orig - old + score);
	  if(ss) ss[count-1] = count-1;
	  /* 	  Rprintf("[%d] %d  %d => %d->%d %.2f\n",count-1,j,i,j,i,moves[count-1].score); */
	  pset_dag_del(ctx,j,i,NULL);
	}
      }
    }
    pset += ctx->rl;
  }
  /*  Rprintf("%d moves\n",count);*/
  return count;
}

double starting_point(pset_dag_ctx ctx) {
  char *vmax = vmaxget();
  char *ext  = (char*)R_alloc(sizeof(char),ctx->n);
  double score = 0.0;
  int i,j;
  pset_dag_clear(ctx);
  for(i=0;i<ctx->n;i++) ext[i] = i;
  for(i=0;i<ctx->n;i++) {
    int a = i+unif_int(ctx->n-i);
    SWAP(ext[i],ext[a]);
    if(i) {
      double S = (1.0-pow(0.5,i));
      double u = unif_rand();
      double C = 0.5;
      for(j=i-1;j>0;j--) 
	if(u < C) break; else { C = C + C*0.5; }
      score += pset_dag_add(ctx,ext[j],ext[i],NULL);
    }
  }
  vmaxset(vmax);
  return score;
}

#define PRIOR(K)   (-ctx->gamma*((double)(K)))

#define RATIO(A,B) ((A+B)?((double)A)/((double)(A+B)):0)

typedef double (*schedfn)(double,double,double,double);

double sched0(double Tn,double T0,double N,double i) {
  return T0-i*((T0-Tn)/N);
}
double sched1(double Tn,double T0,double N,double i) {
  return T0*pow(Tn/T0,i/N);
}
double sched2(double Tn,double T0,double N,double i) {
  double A = (T0-Tn)*((N+1.0)/N);
  double B = T0-A;
  return B+(A/(i+1.0));
}
double sched3(double Tn,double T0,double N,double i) {
  double A = log(T0-Tn)/log(N);
  return T0-pow(i,A);
}
double sched4(double Tn,double T0,double N,double i) {
  return ((T0-Tn)/(1.0+exp(0.3*(i-N/2.0))))+Tn;
}
double sched5(double Tn,double T0,double N,double i) {
  return ((T0-Tn)*(1.0+cos((i*M_PI)/N)))/2 + Tn;
}
double sched6(double Tn,double T0,double N,double i) {
  return ((T0-Tn)*(1.0-tanh((10.0*i)/N - 5)))/2 + Tn;
}
double sched7(double Tn,double T0,double N,double i) {
  return ((T0-Tn)/cosh(10.0*i/N)) + Tn;
}
double sched8(double Tn,double T0,double N,double i) {
  double A = log(T0/Tn)/N;
  return T0*exp(-A*i);
}
double sched9(double Tn,double T0,double N,double i) {
  double A = log(T0/Tn)/(N*N);
  return T0*exp(-A*i*i);
}

#ifdef _N
#undef _N
#endif
#ifdef _T
#undef _T
#endif

SEXP BNSL_FastODAnneal(SEXP d,SEXP max_parents,SEXP _N,SEXP _T,SEXP _sched,SEXP dag,SEXP cand) {
  schedfn sched;
  switch(INTEGER(_sched)[0]) {
  case 1:sched = sched0;break;
  case 2:sched = sched1;break;
  case 3:sched = sched2;break;
  case 4:sched = sched3;break;
  case 5:sched = sched4;break;
  case 6:sched = sched5;break;
  case 7:sched = sched6;break;
  case 8:sched = sched7;break;
  case 9:sched = sched8;break;
  case 10:sched = sched9;break;
  }

  int M = INTEGER(_N)[0];
  int N = INTEGER(_N)[1];
  double T0 = REAL(_T)[0];
  double Tn = REAL(_T)[1];


  Rprintf("%d runs of %d samples\n",M,N);
  int    i,j;
  
  SEXP ret,dags,scores;
  PROTECT(ret = NEW_OBJECT(MAKE_CLASS("sample")));
  SET_SLOT(ret,install("temperature"),duplicate(_T));
  PROTECT(dags = allocVector(VECSXP,M));
  SET_SLOT(ret,install("population"),dags);
  PROTECT(scores = allocVector(REALSXP,M));
  SET_SLOT(ret,install("score"),scores);
  SET_SLOT(ret,install("energy"),ScalarReal(R_NegInf));
  GetRNGstate();
  char *vmax = vmaxget();
  pset_dag_ctx ctx = pset_dag(d,INTEGER(max_parents)[0],0,1);
  move *moves = (move*)R_alloc(sizeof(move),ctx->n*ctx->n*3);
  char *ext   = (char*)R_alloc(sizeof(char),ctx->n);
  double min_seen = R_PosInf;
  for(i=0;i<M;i++) {
    double score,T;
    int    count;
    if(dag == R_NilValue)
      score = -(starting_point(ctx) + PRIOR(ctx->edges));
    else
      score = -(pset_dag_set(ctx,RAW(dag)) + PRIOR(ctx->edges));

    Rprintf("%.2f starting score\n",score);
    for(j=0;j<=N;j++) {
      T = sched(Tn,T0,N,j);
      count = calc_moves(ctx,moves,NULL,-(score+PRIOR(ctx->edges)),cand);
      if(count > 0) {
	move *m = moves + unif_int(count);
	int   edges = ctx->edges;
	switch(m->t) {
	case 1:edges++;break;
	case 2:edges--;break;
	}
	double proposed_score = -(m->score + PRIOR(edges));
	/*	Rprintf("%.2f %.2f %.2f %.2f",proposed_score,score,LogTargetRatio(proposed_score,score,R_NegInf,T),T);*/
	if(unif_rand()    <= exp(LogTargetRatio(proposed_score,score,R_NegInf,T))) {
	  pset_make_move(ctx,m);
	  score = proposed_score;
	  /*	  Rprintf("!\n");*/
	} 
	  /*	  Rprintf("\n");*/
      }
      if(score < min_seen) min_seen = score;
      /*      if(j && j % 15000 == 0)
	      Rprintf("[%d] %d (%.2f) %.2f (min: %.2f)\n",
		i,j,T,score,min_seen);
      */
    }
    Rprintf("[%d] %.2f (min: %.2f)\n",i,score,min_seen);
    SET_VECTOR_ELT(dags,i,pset_make_dag(ctx));
    REAL(scores)[i] = score;
  }
  PutRNGstate();
  vmaxset(vmax);
  UNPROTECT(3);
  return ret;
}

SEXP BNSL_FastOptimize(SEXP d,SEXP max_parents,SEXP _N,SEXP dag,SEXP cand) {

  int M = INTEGER(_N)[0];

  Rprintf("%d optimization runs\n",M);
  int    i,j,k;
  
  SEXP ret,dags,scores;
  PROTECT(ret = NEW_OBJECT(MAKE_CLASS("sample")));
  SET_SLOT(ret,install("temperature"),ScalarReal(1.0));
  PROTECT(dags = allocVector(VECSXP,M));
  SET_SLOT(ret,install("population"),dags);
  PROTECT(scores = allocVector(REALSXP,M));
  SET_SLOT(ret,install("score"),scores);
  SET_SLOT(ret,install("energy"),ScalarReal(R_NegInf));
  GetRNGstate();
  char *vmax = vmaxget();
  pset_dag_ctx ctx = pset_dag(d,INTEGER(max_parents)[0],0,1);
  move *moves = (move*)R_alloc(sizeof(move),ctx->n*ctx->n*3);
  char *ext   = (char*)R_alloc(sizeof(char),ctx->n);
  double min_seen = R_PosInf;
  for(i=0;i<M;i++) {
    double score,T;
    int    count;
    if(dag == R_NilValue)
      score = -(starting_point(ctx) + PRIOR(ctx->edges));
    else
      score = -(pset_dag_set(ctx,RAW(dag)) + PRIOR(ctx->edges));

//    Rprintf("%.2f starting score\n",score);
    int done = 0;
    k = 0;
    while(!done) {
      count = calc_moves(ctx,moves,NULL,-(score+PRIOR(ctx->edges)),cand);
      double best_score = R_PosInf;
      move  *best_move;
      move  *m = moves;
      for(j=0;j<count;j++) {
	int edges = ctx->edges;
	switch(m->t){ 
	case 1:edges++;break;
	case 2:edges--;break;
	}
	double proposed_score = -(m->score + PRIOR(edges));
	/*	Rprintf("[%d %.2f]\n",m->t,proposed_score);*/
	if(proposed_score < best_score) {
	  best_score = proposed_score;
	  best_move  = m;
	}
	m++;
      }
      /*      Rprintf("----\n%.2f %.2f\n",best_score,score);*/
      if(best_score < score) {
	score = best_score;
	pset_make_move(ctx,best_move);
	k++;
      } else done = 1;
    }
//    Rprintf("[%d] %d %.2f\n",i,k,score);
    SET_VECTOR_ELT(dags,i,pset_make_dag(ctx));
    REAL(scores)[i] = score;
  }
  PutRNGstate();
  vmaxset(vmax);
  UNPROTECT(3);
  return ret;
}


extern SEXP BNSL_MakeFastDB(SEXP,SEXP vars,SEXP cache);

SEXP BNSL_SampleLinearExtensions(SEXP dag,SEXP vars,SEXP _N) {
  pset_dag_ctx ctx = pset_dag(BNSL_MakeFastDB(R_NilValue,vars,R_NilValue),
			      0,0,0);
  pset_dag_set(ctx,RAW(dag));

  int i,j,sp = 0,ep = 0;
  char *ext  = (char*)R_alloc(sizeof(char),ctx->n);
  
  //Try not to always get the same starting point for a DAG
  for(i=0;i<ctx->n;i++)
    ext[i] = i;
  for(i=0;i<ctx->n;i++) {
    int a  = i+unif_int(ctx->n-i);
    SWAP(ext[i],ext[a]);
  }
  
  //Do a topological sort
  memcpy(ctx->color,ctx->fanIn,sizeof(char)*ctx->n);
  for(i=0;i<ctx->n;i++)
    if(ctx->fanIn[ext[i]] == 0) ctx->stack[sp++] = ext[i];

  while(sp > 0) {
    char  v   = ctx->stack[--sp];
    char *c   = ctx->child + v*ctx->n;
    ext[ep++] = v;
    for(j=0;j<ctx->fanOut[v];j++) 
      if(ctx->color[c[j]] && --ctx->color[c[j]] == 0) 
	ctx->stack[sp++] = c[j];
  }

  //Whoa, not a DAG?!
  if(ep < ctx->n) error("This is not a DAG!");

  DB *db;
  int unique = 0;
  int count  = 0;
  DBT key,data;
  SEXP ret;
  PROTECT(ret = NEW_OBJECT(MAKE_CLASS("objectSet")));
  db_create(&db,NULL,0);
  db->open(db,NULL,NULL,NULL,DB_BTREE,DB_CREATE,0);
  memset(&key,0,sizeof(DBT));
  memset(&data,0,sizeof(DBT));
  key.data = ext;
  key.size = sizeof(char)*ctx->n;
  int collect = INTEGER(_N)[0];
  int thin    = length(_N)>1?INTEGER(_N)[1]:1;
  int N       = 17*pow(ctx->n,3)*log(ctx->n);
  for(i=0;i<N+(collect*thin);i++) {
    if(unif_rand() > 0.5) {
      int a = unif_int(ctx->n-1);
      if(!bit_test(ctx->adjM + ctx->rl*ext[a+1],ext[a])) {
	SWAP(ext[a],ext[a+1]);
      }
    }
    /* Collect a sample */
    if(i>N && (i-N) % thin == 0) {
      if(db->get(db,NULL,&key,&data,0) == 0) {
	((int*)data.data)[0]++;
      } else {
	count     = 1;
	data.data = &count;
	data.size = sizeof(int);
	unique++;
      }
      db->put(db,NULL,&key,&data,0);
    }
  }
  db->close(db,0);
  UNPROTECT(1);
  return ret;
}



SEXP BNSL_FastODSample(SEXP d,SEXP max_parents,SEXP _N,
		       SEXP dag,
		       SEXP ret,SEXP Q,SEXP X,SEXP S,SEXP cand) {

  GetRNGstate();
  char *vmax = vmaxget();
  /* Set up the context */
  pset_dag_ctx ctx = pset_dag(d,INTEGER(max_parents)[0],0,1);
  int B = INTEGER(_N)[0];
  int N = B+INTEGER(_N)[1];

  int i,j;
  int anneal,anneal_n;
  int accept_jump,reject_jump,accept,reject;

  double T,anneal_a,score,jump_prob,min_seen;

  move *moves = (move*)R_alloc(sizeof(move),ctx->n*ctx->n*3);
  move *moves2= (move*)R_alloc(sizeof(move),ctx->n*ctx->n*3);
  /*  int  *ss    = (int*)R_alloc(sizeof(int),ctx->n*ctx->n*3); */
  char *ext   = (char*)R_alloc(sizeof(char),ctx->n);
  SEXP  scores,dags,start,last;

  ee_ctx ee,je;
  EquiEnergyInit(&ee,ret);
  if(EquiEnergyInit(&je,X) && B > 0) {
    anneal   = 1;
    anneal_n = B/2;
    jump_prob= REAL(Q)[0];
    anneal_a = (log(ee.T)-log(je.T))/((double)anneal_n);
  } else {
    anneal    = 0;
    anneal_n  = 0;
    jump_prob = 0.0;
  }
  T = ee.T;


  accept = reject = accept_jump = reject_jump = 0;
  

  PROTECT(ret    = duplicate(ret));
  PROTECT(scores = allocVector(REALSXP,N));
  SET_SLOT(ret,install("score"),scores);
  PROTECT(dags   = allocVector(VECSXP,N));
  SET_SLOT(ret,install("population"),dags);
  
  /* Initialize our sampler */
  if(dag == R_NilValue)
    score = -(starting_point(ctx) + PRIOR(ctx->edges));
  else 
    score = -(pset_dag_set(ctx,RAW(dag)) + PRIOR(ctx->edges));
  min_seen = score;
  
  Rprintf("p(G)  = %.2f^|G|\n",ctx->gamma);
  Rprintf("pi(x) = exp(-max(h(x),%.2f)/%.2f)\n",ee.H,ee.T);
  Rprintf("jump probability: %.2f, anneal %d,burn-in %d,steps %d\n",
	  jump_prob,anneal_n,B,N-B);
  PROTECT(start = pset_make_dag(ctx));
  last = start;
  for(i=0;i<N;i++) {
    if(anneal & i <= anneal_n)
      T = je.T*exp(anneal_a*(double)i);


    if(X != R_NilValue && unif_rand() < jump_prob) {
      int    ring;SEXP indices;
      double proposed_score;
      for(ring=0;ring<je.chains;ring++) 
	if(Max(score,ee.H) < je.partition[ring]) break;
      indices = VECTOR_ELT(je.rings,ring);
      if(length(indices) > 0) {
	double lr;
	int    index   = unif_int(length(indices));
	index          = INTEGER(indices)[index];
	proposed_score = je.scores[index];
	lr             = LogTargetRatio(proposed_score,score,ee.H,T)
	  + LogTargetRatio(score,proposed_score,je.H,je.T);
	if(lr >= 0.0 || unif_rand() < exp(lr)) {
	  accept_jump++;
	  pset_dag_set(ctx,RAW(VECTOR_ELT(je.population,index)));
	  score = proposed_score;
	} else {
	  reject_jump++;
	}
	goto finish;
      }
    }

    /* If we get this far, make a normal metropolis move */
    {
      int count = calc_moves(ctx,moves,NULL,-(score+PRIOR(ctx->edges)),cand);
      if(count > 0) {
	move  *m  = moves + unif_int(count);
	int    edges = ctx->edges;
	switch(m->t) {
	case 1:edges++;break;
	case 2:edges--;break;
	}
	double proposed_score = -(m->score + PRIOR(edges));
	int count2 = calc_moves(ctx,moves2,NULL,m->score,cand);
	/* We need to do the transition density */
	/*	if(proposed_score < score) Rprintf("%.2f %.2f\n",proposed_score,score);*/
	if(unif_rand()    
	   <= ((double)count2/(double)count)*
	   exp(LogTargetRatio(proposed_score,score,ee.H,T))) {
	  pset_make_move(ctx,m);
	  score = proposed_score;
	  accept++;
	} else {
	  reject++;
	}
      }
    }
    
  finish:
    SET_VECTOR_ELT(dags,i,pset_make_dag(ctx));
    REAL(scores)[i] = score;
    if(score < min_seen) min_seen = score;
    
    /* Display some diagnostics */
    if(i && i % 5000 == 0)
      Rprintf("%d (%.2f) iterations, %.4f low score, %d (%.2f) jumps, %d (%.2f) swaps\n",
	      i,T,min_seen,
	      accept_jump,RATIO(accept_jump,reject_jump),
	      accept,RATIO(accept,reject));
  }
  PutRNGstate();
  ee_ChainDiagnostics(ret,accept_jump,reject_jump,&accept,&reject,1,INTEGER(_N)[0]);
  vmaxset(vmax);
  UNPROTECT(4);
  return ret;
}

#undef PRIOR
#define PRIOR(K)   (-ctx[0]->gamma*((double)(K)))

SEXP BNSL_FastPTSample(SEXP d,SEXP max_parents,SEXP _N,SEXP T,SEXP S,SEXP cand) {
	GetRNGstate();
	char *vmax = vmaxget();
	pset_dag_ctx *ctx   = (pset_dag_ctx*)R_alloc(sizeof(pset_dag_ctx),length(T));
	double       *score = (double*)R_alloc(sizeof(double),length(T));
	double        min_seen;
	int i;
	for(i=0;i<length(T);i++) 
		ctx[i] = pset_dag(d,INTEGER(max_parents)[0],0,1);
	int B = INTEGER(_N)[0];
	int N = B+INTEGER(_N)[1];

	char *temp  =  (char*)R_alloc(sizeof(char),ctx[0]->n*ctx[0]->rl);
	move *moves = (move*)R_alloc(sizeof(move),ctx[0]->n*ctx[0]->n*3);
	move *moves2= (move*)R_alloc(sizeof(move),ctx[0]->n*ctx[0]->n*3);
	char *ext   = (char*)R_alloc(sizeof(char),ctx[0]->n);

	SEXP  ret,scores,dags,start,last;
	PROTECT(ret    = NEW_OBJECT(MAKE_CLASS("sample")));
	PROTECT(scores = allocVector(REALSXP,N));
	SET_SLOT(ret,install("score"),scores);
	PROTECT(dags   = allocVector(VECSXP,N));
	SET_SLOT(ret,install("population"),dags);
  
	/* Initialize our sampler */
	for(i=0;i<length(T);i++)
		score[i] = -(starting_point(ctx[i]) + PRIOR(ctx[i]->edges));
	/* We track on the 1st chain only */
	min_seen = score[0];
  
	Rprintf("STARTING PARALLEL TEMPERING\n");
	Rprintf("p(G)  = %.2f^|G|\n",ctx[0]->gamma);
	PROTECT(start = pset_make_dag(ctx[0]));
	last = start;
	for(i=0;i<N;i++) {
		int j;
		if(unif_rand() < REAL(S)[0]) {
			/* Parallel Tempering */
			j = unif_int(length(T)-1);
			if(unif_rand() < exp(-(score[j+1]-score[j])/REAL(T)[j]+(score[j]-score[j+1])/REAL(T)[j+1])) {
				memcpy(temp,ctx[j]->adjM,ctx[j]->n*ctx[j]->rl);
				score[j] = -(pset_dag_set(ctx[j],ctx[j+1]->adjM) + PRIOR(ctx[j]->edges));
				score[j+1] = -(pset_dag_set(ctx[j+1],temp) + PRIOR(ctx[j+1]->edges));
			}
		} else {
			for(j=0;j<length(T);j++) {
				int count = calc_moves(ctx[j],moves,NULL,-(score[j]+PRIOR(ctx[j]->edges)),cand);
				if(count > 0) {
					move  *m  = moves + unif_int(count);
					int    edges = ctx[j]->edges;
					switch(m->t) {
							case 1:edges++;break;
							case 2:edges--;break;
						}
					double proposed_score = -(m->score + PRIOR(edges));
					int count2 = calc_moves(ctx[j],moves2,NULL,m->score,cand);
					if(unif_rand() <= ((double)count2/(double)count)*exp(LogTargetRatio(proposed_score,score[j],R_NegInf,REAL(T)[j]))) {
						pset_make_move(ctx[j],m);
						score[j] = proposed_score;
					}
				}
			}
		}
		SET_VECTOR_ELT(dags,i,pset_make_dag(ctx[0]));
		REAL(scores)[i] = score[0];
		if(score[0] < min_seen) min_seen = score[0];
		/* Display some diagnostics */
		if(i && i % 5000 == 0)
			Rprintf("%d iterations. %.2f\n",i,min_seen);
	}
	PutRNGstate();
	ee_ChainDiagnostics(ret,0,0,NULL,NULL,0,INTEGER(_N)[0]);
	vmaxset(vmax);
	UNPROTECT(4);
	return ret;
}



typedef struct parents {
  double score;
  char bitset[0];
} parents;

int count_psets(int n,int k) {
  int i,total=0;
  for(i=0;i<=k;i++) total += choose(n,i);
  return total;
}


#if 0
SEXP BNSL_FastCorrectedXXX(SEXP d,SEXP max_parents,SEXP os) {
  GetRNGstate();

  GetRNGstate();
  /* Set up the context */
  pset_dag_ctx ctx = pset_dag(d,INTEGER(max_parents)[0],0,1);
  SEXP orders = GET_SLOT(os,install("objects"));
  SEXP counts = GET_SLOT(os,install("counts"));
  int  o,i,maxpset = 0;
  for(int i=0;i<ctx->n;i++) {
    int count = count_psets(i,ctx->k);
    if(count > maxpset) maxpset = count;
  }
  char *so = (char*)R_alloc(sizeof(char),ctx->n);
  for(i=0;i<ctx->n;i++) so[i] = i;

  int *ndx = (int*)R_alloc(sizeof(int),length(orders));


  Rprintf("%d psets at most will be required\n",maxpset);
  /* Okay, here we go! */
  for(o=0;o<length(counts);o++) {
    int j,N     = INTEGER(counts)[o];
    SEXP  order = VECTOR_ELT(orders,o);
    char *vmax = vmaxget();
    
    /* Random starting position within this order */
    pset_dag_clear(ctx);
    Rprintf("%d ",RAW(order)[0]);
    for(j=1;j<length(order);j++) {
      Rprintf("%d ",RAW(order)[j]);
      pset_dag_random(ctx,RAW(order),j,RAW(order)[j]);
    }
    Rprintf("\n");
    pset_dag_dump(ctx);
    vmaxset(vmax);
  }
  
  PutRNGstate();
  return R_NilValue;
}
#endif
