#include "common.h"
#include "score_search.h"
#include "scoring.h"
#ifndef __APPLE_CC__
#include <bits/nan.h>
#endif

extern SEXP Rf_findVar1(SEXP,SEXP,SEXPTYPE,int);

void init_prior(prior_head p) {
  SEXP x   = Rf_findVar1(install(".Prior"),R_GlobalEnv,REALSXP,FALSE);
  p->prior = (x != R_UnboundValue)?(REAL(x)[0]>0):1;
  
  SEXP b   = Rf_findVar1(install(".Beta"),R_GlobalEnv,REALSXP,FALSE);
  p->beta = (b != R_UnboundValue)?REAL(b)[0]:0.5;


  SEXP g = Rf_findVar1(install(".Gamma"),R_GlobalEnv,REALSXP,FALSE);
  p->gamma = (g != R_UnboundValue)?REAL(g)[0]:1.0;
}

double prior(prior_head p,double m_p,double i) {

  double P = log(p->beta);
  double Q = log(1-p->beta);
  return p->gamma*(m_p*Q + i*(P-Q));
}


/**
 * Without cursor ALARM scoring takes (2.7208,0.665)
 */              

score_ctx bnsl_score_ctx(SEXP d,int k,int mode) {
  int i,j;
  score_ctx ctx = (score_ctx)R_alloc(sizeof(struct score_ctx),1);


  switch(mode) {
  case 1: ctx->fn = logsumexp;break;
  case 2: ctx->fn = findmax;break;
  case 3: ctx->fn = findmin;break;
  default: error("invalid computational mode");
  }

  memset(&(ctx->key),0,sizeof(DBT));
  memset(&(ctx->data),0,sizeof(DBT));
  get_db_info(d,&(ctx->db),(char**)&(ctx->key.data),&(ctx->n));
  ctx->key.size = bitstr_size(ctx->n)+1;
  ctx->db->cursor(ctx->db,NULL,&(ctx->cur),0);

  ctx->k           = k;

  j = ctx->n/2;			/* This is the largest binomial coeff */
  /* Find the largest binomial coefficient allowable by k */
  ctx->score_count = 0;
  for(i=1;i<=ctx->k;i++) {
    int temp = (i > j) ? ctx->n-i : i;
    if(temp > ctx->score_count) ctx->score_count = temp;
  }
  /* Convert this to the largest number of values we will hold */
  /*  printf("C(%d,%d)\n",ctx->n,ctx->score_count);*/
  ctx->score_count = choose(ctx->n,ctx->score_count);

  ctx->scores       = (double*)R_alloc(sizeof(double),ctx->score_count);
  ctx->module_score = (double*)R_alloc(sizeof(double),ctx->k+1);
  ctx->P            = (int*)R_alloc(sizeof(int),ctx->n+2);
  ctx->scratch      = (char*)R_alloc(sizeof(char),ctx->n);
  init_prior((prior_head)ctx);

  return ctx;
}

/*
score_ctx bnsl_score_ctx(SEXP d,int k,int mode) {
  return bnsl_score_ctx_int(d,k,mode,R_alloc);
}

score_ctx bnsl_score_ctx_r(SEXP d,int k,int mode) {
  return bnsl_score_ctx_int(d,k,mode,calloc);
}
*/
void bnsl_score_ctx_done(score_ctx ctx) {
  if(ctx->cur) {
    ctx->cur->c_close(ctx->cur);
    ctx->cur = NULL;
  }
}
void bnsl_score_ctx_done_r(score_ctx ctx) {
  bnsl_score_ctx_done(ctx);
  free(ctx->scores);
  free(ctx->module_score);
  free(ctx->P);
  free(ctx->scratch);
}


#define LOOKUP(X) { \
int r; \
r = ctx->cur->c_get(ctx->cur,&(ctx->key),&(ctx->data),DB_SET); \
if(r == 0) { \
  (X) = ((double*)(ctx->data.data))[0]; \
} else \
  error("%s:%d database lookup error for node (%s)",__FILE__,__LINE__,db_strerror(r)); \
} 
#define LOGADD(A,B) \
  (((A)>(B))?log(1+exp((B)-(A)))+(A):log(1+exp((A)-(B)))+(B))


double bnsl_score_sum_feature(score_ctx ctx,int* o,int o_n,int ext,
			      double *feature) {
  int  i,j,q,t,x,y,z,m_p;
  char   *pa = (char*)(ctx->key.data);
  /* Locate top of feature */
  feature    = feature + (ctx->n*(ext-1));
  
  memset(pa,0,ctx->key.size);
  pa[0] = ext;pa++;
  LOOKUP(ctx->module_score[0]);
  /*  printf("%d {}: %.2f\n",ext,ctx->module_score[0]);*/
  m_p = (o_n < ctx->k) ? o_n : ctx->k;
  for(i=1;i<=m_p;i++) {
    t=0;
    memset(pa,0,bitstr_size(ctx->n));

    inittwiddle(i,o_n,ctx->P);
    /*    printf("%d { ",ext);*/
    for(j=0;j!=o_n-i;j++);
    for(;j!=o_n;j++) {
      /*      Rprintf("%d ",o[j]);*/
      bit_set(pa,o[j]-1);
    }
    LOOKUP(ctx->scores[t++]);
    for(q=0;q<o_n;q++)
      if(bit_test(pa,o[q]-1))
	feature[o[q]-1] = LOGADD(feature[o[q]-1],ctx->scores[t-1]);

    /*    Rprintf("} %.2f\n",ctx->scores[t-1]);*/
    

    while(!twiddle(&x,&y,&z,ctx->P)) {
      bit_set(pa,o[x]-1);
      bit_clear(pa,o[y]-1);
      LOOKUP(ctx->scores[t++]);
      for(q=0;q<o_n;q++)
	if(bit_test(pa,o[q]-1))
	  feature[o[q]-1] = LOGADD(feature[o[q]-1],ctx->scores[t-1]);

      /*      printf("%d { ",ext);*/
      /*      for(j=0;j<o_n;j++)
	      if(bit_test(pa,o[j]-1)) Rprintf("%d ",o[j]);*/
      /*      printf("} %.2f\n",ctx->scores[t-1]);*/
    }
    /*    printf("%d %d\n",ctx->score_count,t);*/
    ctx->module_score[i] = ctx->fn(ctx->scores,t);
  }
  /*  printf("\n");*/
  return ctx->fn(ctx->module_score,m_p+1);
}

double bnsl_score_extension(score_ctx ctx,int* o,int o_n,int ext) {
  int  i,j,t,x,y,z,m_p;
  char *pa = (char*)(ctx->key.data);
  double b=0.0;
  memset(pa,0,ctx->key.size);
  pa[0] = ext;pa++;
  LOOKUP(ctx->module_score[0]);
  /*  printf("%d {}: %.2f\n",ext,ctx->module_score[0]);*/
  m_p = (o_n < ctx->k) ? o_n : ctx->k;
  ctx->module_score[0] += prior((prior_head)ctx,m_p,0);
  for(i=1;i<=m_p;i++) {
    t=0;
    memset(pa,0,bitstr_size(ctx->n));
    b = prior((prior_head)ctx,m_p,i);

    inittwiddle(i,o_n,ctx->P);
    /*    printf("%d { ",ext);*/
    for(j=0;j!=o_n-i;j++);
    for(;j!=o_n;j++) {
      /*      Rprintf("%d ",o[j]);*/
      bit_set(pa,o[j]-1);
    }
    LOOKUP(ctx->scores[t++]);
    ctx->scores[t-1] += b;
    /*    Rprintf("} %.2f\n",ctx->scores[t-1]);*/
    

    while(!twiddle(&x,&y,&z,ctx->P)) {
      bit_set(pa,o[x]-1);
      bit_clear(pa,o[y]-1);
      LOOKUP(ctx->scores[t++]);
      ctx->scores[t-1] += b;
      /*      printf("%d { ",ext);*/
      /*      for(j=0;j<o_n;j++)
	      if(bit_test(pa,o[j]-1)) Rprintf("%d ",o[j]);*/
      /*      printf("} %.2f\n",ctx->scores[t-1]);*/
    }
    /*    printf("%d %d\n",ctx->score_count,t);*/
    ctx->module_score[i] = ctx->fn(ctx->scores,t);
  }
  /*  printf("\n");*/
  return ctx->fn(ctx->module_score,m_p+1);
}

double bnsl_score_ordering_features(score_ctx ctx,int *o,int o_n,
				    double*features) {
  int i,j;
  char *vmax = vmaxget();
  double total,total_score = 0;
  for(i=0;i<o_n*o_n;i++) {
    features[i] = R_NegInf;
  }
  for(i=1;i<o_n;i++) {
    total = bnsl_score_sum_feature(ctx,o,i,o[i],features);
    for(j=0;j<i;j++) {
      features[((o[i]-1)*o_n)+(o[j]-1)] -= total;
    }
    total_score += total;
  }
  return total_score;
}


double bnsl_score_ordering(score_ctx ctx,int*o,int o_n) {
  int i;
  double total_score = 0.0;
  for(i=0;i<o_n;i++)
    total_score += bnsl_score_extension(ctx,o,i,o[i]);
  return total_score;
}

double bnsl_score_ordering_update(score_ctx ctx,int*o,int o_n,int a,int b) {
  int i;
  double total_score = 0.0;
  for(i=a;i<=b;i++) total_score += bnsl_score_extension(ctx,o,i,o[i]);
  return total_score;
}

SEXP BNSL_FeatureProbs(SEXP d,SEXP max_parents,SEXP orders,SEXP mode) {
  int  i,j,k,a,b;
  SEXP featmat,dim;
  char *vmax    = vmaxget();
  score_ctx ctx = bnsl_score_ctx(d,INTEGER(max_parents)[0],INTEGER(mode)[0]);
  PROTECT(featmat = allocVector(REALSXP,ctx->n*ctx->n*length(orders)));
  PROTECT(dim     = allocVector(REALSXP,2));
  REAL(dim)[0] = (double)(ctx->n*ctx->n);
  REAL(dim)[1] = (double)length(orders);
  memset(REAL(featmat),0,sizeof(double)*(ctx->n*ctx->n*length(orders)));
  for(i=0;i<length(orders);i++) {
    SEXP    order = VECTOR_ELT(orders,i);
    double *feat  = REAL(featmat) + i*(ctx->n*ctx->n);
    Rprintf("%d: ",length(order));
    for(j=0;j<length(order);j++)
      Rprintf("%d ",INTEGER(order)[j]);
    Rprintf("\n");
    bnsl_score_ordering_features(ctx,INTEGER(order),ctx->n,feat);
    for(j=0;j<ctx->n*ctx->n;j++) feat[j] = exp(feat[j]);
  }
  setAttrib(featmat,R_DimSymbol,dim);
  UNPROTECT(2);
  return featmat;
}


SEXP BNSL_ScoreFeatures(SEXP d,SEXP max_parents,SEXP orders,SEXP mode,
			SEXP weight) {
  SEXP      feat;
  char     *vmax = vmaxget();
  score_ctx ctx = bnsl_score_ctx(d,INTEGER(max_parents)[0],INTEGER(mode)[0]);
  int       i,j,k,n   = length(orders);
  int       l         = (weight != R_NilValue)?length(weight):0;
  double   *list= (double*)R_alloc(sizeof(double),n*(ctx->n*ctx->n));
  double    total;
  if(TYPEOF(orders) != VECSXP) 
    error("Orders must be a list of vectors");
  PROTECT(feat = allocVector(REALSXP,ctx->n*ctx->n));
  for(i=0;i<n;i++) {
    SEXP order = VECTOR_ELT(orders,i);
    if(i%1000==0 && i) Rprintf(".");
    if(TYPEOF(order) != INTSXP) error("Invalid order type");
    total = bnsl_score_ordering_features(ctx,INTEGER(order),ctx->n,REAL(feat));
    for(j=0;j<ctx->n*ctx->n;j++) {
      list[j*n + i] = 
	(REAL(feat)[j] != R_NegInf)?
	(REAL(feat)[j]+(TYPEOF(weight)==REALSXP?REAL(weight)[i%l]:0.0)):
	R_NegInf;
    }
  }
  for(i=0;i<ctx->n*ctx->n;i++) {
    REAL(feat)[i] = logsumexp(&list[i*n],n);
    if(TYPEOF(weight)!=REALSXP) REAL(feat)[i] -= log((double)n);
  }

  vmaxset(vmax);
  UNPROTECT(1);
  return feat;
}

SEXP BNSL_ScoreOrder(SEXP d,SEXP order,SEXP max_parents,SEXP mode) {
  SEXP res;
  score_ctx ctx = bnsl_score_ctx(d,INTEGER(max_parents)[0],INTEGER(mode)[0]);
  res = ScalarReal(bnsl_score_ordering(ctx,INTEGER(order),length(order)));
  bnsl_score_ctx_done(ctx);
  return res;
}

pset_ctx bnsl_pset_ctx(SEXP d,int k) {
  int i,j,sc;
  pset_ctx ctx = (pset_ctx)R_alloc(sizeof(struct pset_ctx),1);
  init_prior((prior_head)ctx);
  memset(&(ctx->key),0,sizeof(DBT));
  memset(&(ctx->data),0,sizeof(DBT));
  get_db_info(d,&(ctx->db),(char**)&(ctx->key.data),&(ctx->n));
  ctx->key.size = bitstr_size(ctx->n)+1;
  ctx->db->cursor(ctx->db,NULL,&(ctx->cur),0);

  ctx->k = k;
  j      = ctx->n/2; /* This is the largest binomial coeff */
  /* Find the largest binomial coefficient allowable by k  */
  ctx->score_count = 0;
  for(i=1;i<=ctx->k;i++) {
    int temp = (i > j) ? ctx->n-i : i;
    if(temp > ctx->score_count) ctx->score_count = temp;
  }
  sc = 1;
  for(i=1;i<=ctx->score_count;i++)
    sc += choose(ctx->n,i);
  ctx->score_count = sc;
  ctx->scores  = (double*)R_alloc(sizeof(double),ctx->score_count);
  ctx->order   = (int*)   R_alloc(sizeof(int),ctx->score_count);
  ctx->pset    = (char*)  R_alloc(sizeof(char),
				  bitstr_size(ctx->n)*ctx->score_count);
  ctx->P       = (int*)   R_alloc(sizeof(int),ctx->n+2);
  ctx->scratch = (char*)  R_alloc(sizeof(char),ctx->n);

  return ctx;
}

void bnsl_pset_ctx_done(pset_ctx ctx) {
  if(ctx->cur) {
    ctx->cur->c_close(ctx->cur);
    ctx->cur = NULL;
  }
}

void bnsl_pset_set_vertex(pset_ctx ps,int v) {
  ps->v = v;
  char *pa = (char*)(ps->key.data);
  pa[0] = v;
  pa++;
  memset(pa,0,bitstr_size(ps->n));
}
void bnsl_pset_set_parent(pset_ctx ps,int p) {
  char *pa = (char*)(ps->key.data);
  pa++;
  bit_set(pa,p-1);
}
void bnsl_pset_remove_parent(pset_ctx ps,int p) {
  char *pa = (char*)(ps->key.data);
  pa++;
  bit_clear(pa,p-1);  
}

double bnsl_pset_get_score(pset_ctx ctx) {
  double score;
  LOOKUP(score);
  return score;
}

void bnsl_pset_dump(pset_ctx ps,int off) {
  char *pa = ps->pset + bitstr_size(ps->n)*ps->order[off];
  int   i;
  Rprintf("(");
  for(i=0;i<ps->n;i++)
    if(bit_test(pa,i))
      Rprintf("%d ",i+1);
  Rprintf("->%d)",ps->v);
}


#define DRAWPA { \
    int x; \
    for(x=0;x<ctx->n;x++) \
      Rprintf(bit_test(pa,x)?"1":"0"); \
  }

double bnsl_pset_prepare(pset_ctx ctx,int *o,int o_n,int ext) {
  int i,j,t,score,x,y,z,m_p,spos;
  char *pa = (char*)(ctx->key.data);
  memset(ctx->scores,0,sizeof(double)*ctx->score_count);
  memset(ctx->order,0,sizeof(int)*ctx->score_count);
  memset(pa,0,ctx->key.size);
  score = t = 0;
  ctx->v=ext;
  pa[0]=ext;pa++;
  LOOKUP(ctx->scores[score++]);
  ctx->scores[score-1] += prior((prior_head)ctx,m_p,i);
  ctx->order[t++] = score-1;
  memcpy(ctx->pset + bitstr_size(ctx->n)*(score-1),pa,bitstr_size(ctx->n));
  m_p = (o_n < ctx->k) ? o_n : ctx->k;
  for(i=1;i<=m_p;i++) {
    double b = prior((prior_head)ctx,m_p,i);
    memset(pa,0,bitstr_size(ctx->n));
    inittwiddle(i,o_n,ctx->P);
    for(j=0;j!=o_n-i;j++);
    for(;j!=o_n;j++) bit_set(pa,o[j]-1);
    LOOKUP(ctx->scores[score++]);
    ctx->scores[score-1] += b;
    ctx->order[t++] = score-1;
    memcpy(ctx->pset + bitstr_size(ctx->n)*(score-1),pa,bitstr_size(ctx->n));
    while(!twiddle(&x,&y,&z,ctx->P)) {
      bit_set(pa,o[x]-1);
      bit_clear(pa,o[y]-1);
      LOOKUP(ctx->scores[score++]);
      ctx->scores[score-1] += b;
      ctx->order[t++] = score-1;
      memcpy(ctx->pset + bitstr_size(ctx->n)*(score-1),
	     pa,bitstr_size(ctx->n));      
    }
  }
  ctx->used = score;
  revsort(ctx->scores,ctx->order,score);
  ctx->norm = logsumexp(ctx->scores,score);
  /*  Rprintf("%d:",ctx->v);*/

  return ctx->norm;
}

double bnsl_pset_prob(pset_ctx ctx,int *o,int o_n,int ext,char *pset) {
  int   i,j,t,score,x,y,z,m_p,spos;
  double norm,p=0.0;
  char *pa = (char*)(ctx->key.data);

  memset(ctx->scores,0,sizeof(double)*ctx->score_count);
  memset(ctx->order,0,sizeof(int)*ctx->score_count);
  
  memset(pa,0,ctx->key.size);
  pa[0] = ext;pa++;
  memcpy(pa,pset,bitstr_size(ctx->n));
  LOOKUP(p);
  memset(pa,0,bitstr_size(ctx->n));
  score = t = 0;
  LOOKUP(ctx->scores[score++]);
  m_p = (o_n < ctx->k) ? o_n : ctx->k;
  ctx->scores[score-1] += prior((prior_head)ctx,m_p,0);
  for(i=1;i<=m_p;i++) {
    double b = prior((prior_head)ctx,m_p,i);
    memset(pa,0,bitstr_size(ctx->n));
    inittwiddle(i,o_n,ctx->P);
    for(j=0;j!=o_n-i;j++);
    for(;j!=o_n;j++) bit_set(pa,o[j]-1);
    LOOKUP(ctx->scores[score++]);
    ctx->scores[score-1] += b;
    while(!twiddle(&x,&y,&z,ctx->P)) {
      bit_set(pa,o[x]-1);
      bit_clear(pa,o[y]-1);
      LOOKUP(ctx->scores[score++]);
      ctx->scores[score-1] += b;
    }
  }
  norm = logsumexp(ctx->scores,score);
  return p-norm;
}

double bnsl_pset_sample(pset_ctx ctx,int *o,int o_n,int ext,
			char* pset) {
  int   i,j,t,score,x,y,z,m_p,spos;
  double norm,cdf,ru,p;
  char *pa = (char*)(ctx->key.data);

  memset(ctx->scores,0,sizeof(double)*ctx->score_count);
  memset(ctx->order,0,sizeof(int)*ctx->score_count);
  
  memset(pa,0,ctx->key.size);
  pa[0] = ext;pa++;
  score = t = 0;
  LOOKUP(ctx->scores[score++]);
  ctx->order[t++] = score;
  m_p = (o_n < ctx->k) ? o_n : ctx->k;
  for(i=1;i<=m_p;i++) {
    memset(pa,0,bitstr_size(ctx->n));
    inittwiddle(i,o_n,ctx->P);
    for(j=0;j!=o_n-i;j++);
    for(;j!=o_n;j++) bit_set(pa,o[j]-1);
    LOOKUP(ctx->scores[score++]);
    ctx->order[t++] = score;
    while(!twiddle(&x,&y,&z,ctx->P)) {
      bit_set(pa,o[x]-1);
      bit_clear(pa,o[y]-1);
      LOOKUP(ctx->scores[score++]);
      ctx->order[t++] = score;
    }
  }

  /* Calculate the normalizing constant and sample a parent set*/
  norm = logsumexp(ctx->scores,score);
  revsort(ctx->scores,ctx->order,score);
  cdf  = 0.0;ru = unif_rand();
  for(i=0;i<score;i++) {
    cdf += exp(ctx->scores[i]-norm);
    spos = ctx->order[i];
    if(ru < cdf) break;
  }
  p = exp(ctx->scores[i]-norm);
  /*  Rprintf("%d %.2f\n",spos,p);*/
  

  /* Find the right parent set */
  if(pset) {
    memset(pset,0,bitstr_size(ctx->n));
    t=0;
    if(++t == spos) return p;
    for(i=1;i<=m_p;i++) {
      memset(pset,0,bitstr_size(ctx->n));
      inittwiddle(i,o_n,ctx->P);
      for(j=0;j!=o_n-i;j++);
      for(;j!=o_n;j++)
	bit_set(pset,o[j]-1);
      if(++t == spos) return p;
      while(!twiddle(&x,&y,&z,ctx->P)) {
	bit_set(pset,o[x]-1);
	bit_clear(pset,o[y]-1);
	if(++t == spos) return p;
      }
    }
  }
  return p;
}

double bnsl_pset_sample_graph(pset_ctx ctx,dag_ctx dag,
			      int*o,int o_n) {
  int i;
  double p;
  bnsl_dag_clear(dag);
  dag->scores[0]   = p = 1.0;
  char *pset = R_alloc(sizeof(char),bitstr_size(ctx->n));
  for(i=1;i<o_n;i++) {
    int j;
    dag->scores[i] = bnsl_pset_sample(ctx,o,i,o[i],pset);
    p *= dag->scores[i];
    for(j=0;j<ctx->n;j++)
      if(bit_test(pset,o[j]-1)) {
	bnsl_dag_add_edge(dag,o[j],o[i]);
      }
  }
  return p;
}

void bnsl_dag_clear(dag_ctx ctx) {
  int i;
  memset(ctx->dag,0,ctx->key.size*ctx->n);
  memset(ctx->rdag,0,ctx->key.size*ctx->n);
  memset(ctx->degree,0,sizeof(int)*ctx->n);
  memset(ctx->scores,0,sizeof(double)*ctx->n);
  ctx->edges=0;
  for(i=0;i<ctx->n;i++) {
    ctx->dag[i*ctx->key.size] = (unsigned char)i+1;
    ctx->rdag[i*ctx->key.size]= (unsigned char)i+1;
  }
}
void bnsl_dag_clear_vertex(dag_ctx ctx,int vertex) {
  int i;
  /* Remove all parent edges for this vertex */
  for(i=0;i<ctx->n;i++) bnsl_dag_set_edge_no(ctx,i+1,vertex,0);
}


void bnsl_dag_set_pset(dag_ctx ctx,pset_ctx ps,int off) {
  char *pa     = ps->pset + bitstr_size(ps->n)*ps->order[off];
  int   i;
  /* Get the parent set offset */
  for(i=0;i<ps->n;i++)
    bnsl_dag_set_edge_no(ctx,i+1,ps->v,bit_test(pa,i)?1:0);
  bnsl_dag_update_score(ctx,ps->v);
}

void bnsl_dag_copy(dag_ctx dest,dag_ctx src) {
  memcpy(dest->dag,src->dag,src->key.size*src->n);
  memcpy(dest->rdag,src->rdag,src->key.size*src->n);
  memcpy(dest->degree,src->degree,sizeof(int)*src->n);
  memcpy(dest->scores,src->scores,sizeof(double)*src->n);
  dest->edges = src->edges;
}

dag_ctx bnsl_dag_ctx(SEXP d,int k) {
  int i,j;
  dag_ctx ctx = (dag_ctx)R_alloc(sizeof(struct dag_ctx),1);
  memset(&(ctx->key),0,sizeof(DBT));
  memset(&(ctx->data),0,sizeof(DBT));
  get_db_info(d,&(ctx->db),(char**)&(ctx->key.data),&(ctx->n));
  ctx->k = k;
  ctx->key.size = bitstr_size(ctx->n)+1;
  ctx->db->cursor(ctx->db,NULL,&(ctx->cur),0);
  
  /* Allocate an initialize the DAG */
  ctx->dag = (char*)R_alloc(ctx->key.size,ctx->n);
  ctx->rdag= (char*)R_alloc(ctx->key.size,ctx->n);
  memset(ctx->dag,0,ctx->key.size*ctx->n);
  memset(ctx->rdag,0,ctx->key.size*ctx->n);
  
  ctx->degree = (int*)R_alloc(sizeof(int),ctx->n);
  memset(ctx->degree,0,sizeof(int)*ctx->n);
  ctx->scores = (double*)R_alloc(sizeof(double),ctx->n);
  memset(ctx->scores,0,sizeof(double)*ctx->n);

  ctx->edges=0;
  for(i=0;i<ctx->n;i++) {
    ctx->dag[i*ctx->key.size] = (unsigned char)i+1;
    ctx->rdag[i*ctx->key.size] = (unsigned char)i+1;
  }
  return ctx;
}
void bnsl_dag_ctx_done(dag_ctx ctx) {
  if(ctx->cur) {
    ctx->cur->c_close(ctx->cur);
    ctx->cur = NULL;
  }
}
double bnsl_dag_update_score(dag_ctx ctx,int to) {
  int r,i;
  ctx->key.data = &(ctx->dag[((to-1)*ctx->key.size)]);
  r=ctx->db->get(ctx->db,NULL,&(ctx->key),&(ctx->data),0);
  if(r==0)
    ctx->scores[to-1] =  ((double*)(ctx->data.data))[0];
  else {
    bnsl_dag_dump(ctx);
    error("%s:%d database lookup error for [%d] (%s)",
	  __FILE__,__LINE__,to,db_strerror(r));
  }
  return ctx->scores[to-1];
}

double bnsl_dag_get_score(dag_ctx ctx,int to) { 
  double x = ctx->scores[to-1];
  return x;
}
double bnsl_dag_score(dag_ctx ctx) {
  int i;
  double score=0.0;
  for(i=0;i<ctx->n;i++) score += ctx->scores[i];
  return score;
}

int bnsl_dag_has_cycle(dag_ctx ctx,int start) {
  int i,stack=0;
  char* vmax = vmaxget();
  int *edge = (int*)R_alloc(sizeof(int),ctx->edges);
  int *color= (int*)R_alloc(sizeof(int),ctx->n);
  memset(color,0,sizeof(int)*ctx->n);
  edge[stack++] = start;
  do {
    int working = edge[--stack];
    char *p     = &(ctx->rdag[(working*ctx->key.size)+1]);
    if(++color[working] == 2 && working == start) {
      vmaxset(vmax);
      return 1; /* We got back to where we started. Damn! */
    }
    for(i=0;i<ctx->n;i++)
      if(bit_test(p,i)) edge[stack++] = i;
  } while(stack>0);
  vmaxset(vmax);
  return 0;
}

int bnsl_dag_connected(dag_ctx ctx,int start) {
  int head=0,tail=0,i;
  char *vmax = vmaxget();
  int*  edge = (int*)R_alloc(sizeof(int),ctx->edges);
  int*  color= (int*)R_alloc(sizeof(int),ctx->n);
  int   compcount=0;
  memset(color,0,sizeof(int)*ctx->n);
  edge[tail++] = start;
  do {
    int working = edge[head++];
    char *p     = &(ctx->dag[(working*ctx->key.size)+1]);
    char *rp    = &(ctx->rdag[(working*ctx->key.size)+1]);
    if(color[working]==0) {
      color[working]=1;
      for(i=0;i<ctx->n;i++) {
	if(bit_test(p,i)  && color[i]==0) edge[tail++] = i;
	if(bit_test(rp,i) && color[i]==0) edge[tail++] = i;
      }
      compcount++;
    }
  } while(head < tail);
  return (compcount == ctx->n) ? 1 : 0;
}

void bnsl_dag_dump_vertex(dag_ctx ctx,int vertex) {
  int i = vertex-1,j;
  char *p = &(ctx->dag[i*ctx->key.size]);
  int   m = p[0];p++;
  for(j=0;j<ctx->n;j++)
    if(bit_test(p,j)) Rprintf(" %d",j+1);
  Rprintf("->%d\n",m);
}

void bnsl_dag_dump(dag_ctx ctx) {
  int i,j;
  for(i=0;i<ctx->n;i++) {
    char *p = &(ctx->dag[i*ctx->key.size]);
    int   m = p[0];p++;
    for(j=0;j<ctx->n;j++)
      if(bit_test(p,j)) Rprintf(" %d",j+1);
    Rprintf("->%d (%.2f)\n",m,ctx->scores[i]);
  }
}

char* bnsl_dag_parent_set(dag_ctx ctx,int j) {
  char *p = &(ctx->dag[(j-1)*ctx->key.size]);
  int   m = p[0];p++;
  if(j != m) error("dag incorrectly configured");
  return p;
}
char* bnsl_dag_child_set(dag_ctx ctx,int j) {
  char *p = &(ctx->dag[(j-1)*ctx->key.size]);
  int   m = p[0];p++;
  if(j != m) error("dag incorrectly configured");
  return p;
}

void bnsl_dag_set_edge_no(dag_ctx ctx,int from,int to,int state) {
  char*p = &(ctx->dag[((to-1)*ctx->key.size)+1]);
  char*rp= &(ctx->rdag[((from-1)*ctx->key.size)+1]);
  if(state) {
    if(!bit_test(p,from-1)) {
      bit_set(p,from-1);
      bit_set(rp,to-1);
      ctx->edges++;
      ctx->degree[to-1]++;
    }
  } else {
    if(bit_test(p,from-1)) {
      bit_clear(p,from-1);
      bit_clear(p,to-1);
      ctx->edges--;
      ctx->degree[to-1]--;
    }
  }
}

double bnsl_dag_set_edge(dag_ctx ctx,int from,int to,int state) {
  char*p = &(ctx->dag[((to-1)*ctx->key.size)+1]);
  char*rp= &(ctx->rdag[((from-1)*ctx->key.size)+1]);
  if(state) {
    if(!bit_test(p,from-1)) {
      bit_set(p,from-1);
      bit_set(rp,to-1);
      ctx->edges++;
      ctx->degree[to-1]++;
    } else Rprintf("(%d->%d)?",from,to);
  } else {
    if(bit_test(p,from-1)) {
      bit_clear(p,from-1);
      bit_clear(p,to-1);
      ctx->edges--;
      ctx->degree[to-1]--;
    } else Rprintf("(%d->%d)/",from,to);
  }
  bnsl_dag_update_score(ctx,to);
  return bnsl_dag_get_score(ctx,to);
}

int bnsl_dag_add_edge(dag_ctx ctx,int from,int to) {
  char*p = &(ctx->dag[((to-1)*ctx->key.size)+1]);
  char*rp= &(ctx->rdag[((from-1)*ctx->key.size)+1]);
  /*  Rprintf("add %d -> %d",from,to); */
  if(bit_test(p,from-1)) return 0;
  if(ctx->degree[to-1] == ctx->k) return 0;
  bit_set(p,from-1);
  bit_set(rp,to-1);
  ctx->edges++;			/* This needs to be here for cycle checks */
  if(bnsl_dag_has_cycle(ctx,to-1)) {
    bit_clear(p,from-1);
    bit_clear(rp,to-1);
    ctx->edges--;
    /*    Rprintf(" failed\n"); */
    return 0;
  } else {

    ctx->degree[to-1]++;
  }
  /*  Rprintf(" succeeded\n"); */
  return 1;
}

int bnsl_dag_reverse_edge(dag_ctx ctx,int from,int to) {
  if(!bnsl_dag_remove_edge(ctx,from,to)) return 0;
  if(!bnsl_dag_add_edge(ctx,to,from)) {
    bnsl_dag_add_edge(ctx,from,to);
    return 0;
  }
  return 1;
}


int bnsl_dag_remove_edge(dag_ctx ctx,int from,int to) {
  char*p = &(ctx->dag[((to-1)*ctx->key.size)+1]);
  char*rp= &(ctx->rdag[((from-1)*ctx->key.size)+1]);
  /*  Rprintf("remove %d -> %d",from,to); */
  /* Can't remove non-existant edges */
  if(!bit_test(p,from-1)) return 0;
  
  bit_clear(p,from-1);
  bit_clear(rp,to-1);
  ctx->edges--;
  /*  if(!bnsl_dag_connected(ctx,to-1)) {
      bit_set(p,from-1);
      bit_set(rp,to-1);
    ctx->edges++;
    return 0;
    } else */
  ctx->degree[to-1]--;
  return 1;
}

int bnsl_dag_remove_edge_nocheck(dag_ctx ctx,int from,int to) {
  char*p = &(ctx->dag[((to-1)*ctx->key.size)+1]);
  char*rp= &(ctx->rdag[((from-1)*ctx->key.size)+1]);
  /*  Rprintf("remove %d -> %d",from,to); */
  bit_clear(p,from-1);
  bit_clear(rp,to-1);
  ctx->edges--;
  ctx->degree[to-1]--;
  /*  Rprintf(" succeeded\n");*/
  return 1;
}

int bnsl_dag_has_edge(dag_ctx ctx,int from,int to) {
  char*p = &(ctx->dag[((to-1)*ctx->key.size)+1]);
  return (bit_test(p,from-1)) ? 1 : 0;
}

int bnsl_dag_degree(dag_ctx ctx,int edge) { return ctx->degree[edge-1]; }
