#ifndef _h_scoring_
#define _h_scoring_

typedef struct score_ctx *score_ctx;
typedef struct pset_ctx  *pset_ctx;
typedef struct dag_ctx   *dag_ctx;
typedef struct prior_head *prior_head;

struct prior_head {
  int    prior;
  double beta;
  double gamma;
};

void init_prior(prior_head p);
double prior(prior_head,double,double);

struct score_ctx {
  struct prior_head prior;
  DB  *db;
  DBC *cur;
  DBT key,data;
  
  int     score_count;
  int*    P;
  char*   scratch;
  double* scores;		/* Number of scores-per-module */
  double* module_score;		/* Number of modules */
  int     n,k;			/* n = vars,k = max parents */
  double  (*fn)(double*,int);
};

score_ctx bnsl_score_ctx(SEXP d,int k,int mode);
score_ctx bnsl_score_ctx_r(SEXP d,int k,int mode);
void   bnsl_score_ctx_done(score_ctx ctx);
double bnsl_score_extension(score_ctx ctx,int* o,int o_n,int ext);
double bnsl_score_ordering(score_ctx ctx,int*o,int o_n);
double bnsl_score_ordering_update(score_ctx ctx,int*o,int o_n,int a,int b);
double bnsl_score_sum_feature(score_ctx ctx,int* o,int o_n,int ext,
			      double *feature);
double bnsl_score_ordering_features(score_ctx ctx,int *o,int o_n,
				    double*features);


struct pset_ctx {
  struct prior_head prior;
  DB*     db;
  DBC*    cur;
  DBT     key,data;

  int     v;			/* My vertex */
  char*   scratch;
  int*    P;
  int     score_count;		/* The largest set we can hold */
  int     used;			/* The number of scores employed */
  double* scores;		/* Score list */
  double   norm;		/* Normalizing constant */
  int*    order;		/* Ordering   */
  char*   pset;		/* The parent sets */
  int     n,k;			/* The max parents */
};

pset_ctx bnsl_pset_ctx(SEXP d,int k);
void     bnsl_pset_ctx_done(pset_ctx);
double   bnsl_pset_prepare(pset_ctx ctx,int*o,int o_n,int ext);
double   bnsl_pset_sample(pset_ctx ctx,int*o,int o_n,int ext,char*pset);
double   bnsl_pset_sample_graph(pset_ctx ctx,dag_ctx dag,int *o,int o_n);
double   bnsl_pset_prob(pset_ctx ctx,int*o,int o_n,int ext,char*pset);
void     bnsl_pset_dump(pset_ctx ctx,int off);

void     bnsl_pset_set_vertex(pset_ctx,int v);
void     bnsl_pset_set_parent(pset_ctx,int e);
void     bnsl_pset_remove_parent(pset_ctx,int e);
double   bnsl_pset_get_score(pset_ctx);


struct dag_ctx {
  DB  *db;
  DBC *cur;
  DBT  key,data;


  int     edges;		/* The number of edges in the graph */

  char*   dag;			/* Packed DAG representation */
  char*   rdag;			/* Reverse of the packed DAG */
  int*    degree;		/* The current degree of each node */
  double* scores;		/* Scores for the particular configuration */
  int     n,k;			/* n = vars, k = max parents */
};

dag_ctx bnsl_dag_ctx(SEXP d,int k);
void    bnsl_dag_ctx_done(dag_ctx ctx);
void    bnsl_dag_copy(dag_ctx dest,dag_ctx src);
void    bnsl_dag_clear(dag_ctx ctx);
void    bnsl_dag_clear_vertex(dag_ctx,int);
double  bnsl_dag_score(dag_ctx ctx);
void    bnsl_dag_dump_vertex(dag_ctx ctx,int vertex);
void    bnsl_dag_dump(dag_ctx ctx);
double  bnsl_dag_get_score(dag_ctx ctx,int to);
double  bnsl_dag_update_score(dag_ctx,int);
char*   bnsl_dag_parent_set(dag_ctx,int);
char*   bnsl_dag_child_set(dag_ctx,int);
double  bnsl_dag_set_edge(dag_ctx,int from,int to,int set);
void    bnsl_dag_set_edge_no(dag_ctx,int from,int to,int set);
void    bnsl_dag_set_pset(dag_ctx,pset_ctx,int x);
int     bnsl_dag_degree(dag_ctx,int);
int     bnsl_dag_has_edge(dag_ctx,int,int);

int     bnsl_dag_has_cycle(dag_ctx ctx,int start);
int     bnsl_dag_connected(dag_ctx ctx,int start);
int     bnsl_dag_add_edge(dag_ctx ctx,int i,int j);
int     bnsl_dag_reverse_edge(dag_ctx ctx,int i,int j);
int     bnsl_dag_remove_edge(dag_ctx ctx,int i,int j);
int     bnsl_dag_remove_edge_nocheck(dag_ctx ctx,int i,int j);


#endif
