#include "common.h"
#include "score_search.h"
#include "scoring.h"

SEXP BNSL_SequentialResample(SEXP W,SEXP d,SEXP max_parents,SEXP mode,
			     SEXP temp,SEXP count) {
  SEXP ret,names,orders,weights,scores;
  /* Get a scoring context */
  score_ctx ctx = bnsl_score_ctx(d,INTEGER(max_parents)[0],
				 INTEGER(mode)[0]);
  double T = REAL(temp)[0],invT=1.0/T;
  double*w = REAL(W),*pdf;
  int i,j,k,l;
  
  PROTECT(ret    = allocVector(VECSXP,3));
  PROTECT(names  = allocVector(STRSXP,3));
  PROTECT(orders = allocVector(VECSXP,INTEGER(count)[0]));
  PROTECT(weights= allocVector(REALSXP,INTEGER(count)[0]));
  PROTECT(scores = allocVector(REALSXP,INTEGER(count)[0]));
  SET_STRING_ELT(names,0,mkChar("ordering"));SET_VECTOR_ELT(ret,0,orders);
  SET_STRING_ELT(names,1,mkChar("score"));SET_VECTOR_ELT(ret,1,scores);
  SET_STRING_ELT(names,2,mkChar("weight"));SET_VECTOR_ELT(ret,2,weights);
  setAttrib(ret,R_NamesSymbol,names);
  setAttrib(ret,install("temperature"),ScalarReal(T));


  
  for(i=0;i<INTEGER(count)[0];i++) {
    SEXP order;
    PROTECT(order = allocVector(INTSXP,ctx->n));

    /* Initialize the ordering */
    REAL(weights)[i] = 0.0;
    for(j=0;j<ctx->n;j++) INTEGER(order)[j] = j+1;


    for(j=0;j<ctx->n-1;j++) {
      double* pdf  = &w[j*ctx->n];
      double  norm = 0.0,cdf=0.0,ru = unif_rand();
      int     swap;
      for(k=j;k<ctx->n;k++) norm += pdf[INTEGER(order)[k]-1];
      if(norm) {
	for(k=j;k<ctx->n;k++) {
	  cdf += pdf[INTEGER(order)[k]-1]/norm;
	  if(ru < cdf) break;
	}
	REAL(weights)[i] += log(pdf[INTEGER(order)[k]-1]/norm);
	swap = INTEGER(order)[j];
	INTEGER(order)[j] = INTEGER(order)[k];
	INTEGER(order)[k] = swap;
      }
    }
    REAL(scores)[i] = invT*bnsl_score_ordering(ctx,INTEGER(order),ctx->n);
    SET_VECTOR_ELT(orders,i,order);
    UNPROTECT(1);
  }
  bnsl_score_ctx_done(ctx);
  UNPROTECT(5);
  return ret;
}

SEXP BNSL_SequentialSamplerStart(SEXP d,SEXP max_parents,SEXP mode,
				 SEXP temp,SEXP count) {
  /* Get a scoring context */
  score_ctx ctx = bnsl_score_ctx(d,INTEGER(max_parents)[0],
				 INTEGER(mode)[0]);
  /* Return values */
  SEXP orders,scores,weights;
  SEXP ret,names;

  int i;
  double T = REAL(temp)[0],invT = 1.0/T;
  double*pdf,norm;

  /* Set up the return block */
  PROTECT(ret    = allocVector(VECSXP,3));
  PROTECT(names  = allocVector(STRSXP,3));
  PROTECT(orders = allocVector(VECSXP,INTEGER(count)[0]));
  PROTECT(weights= allocVector(REALSXP,INTEGER(count)[0]));
  PROTECT(scores = allocVector(REALSXP,INTEGER(count)[0]));
  SET_STRING_ELT(names,0,mkChar("ordering"));SET_VECTOR_ELT(ret,0,orders);
  SET_STRING_ELT(names,1,mkChar("score"));SET_VECTOR_ELT(ret,1,scores);
  SET_STRING_ELT(names,2,mkChar("weight"));SET_VECTOR_ELT(ret,2,weights);
  setAttrib(ret,R_NamesSymbol,names);
  setAttrib(ret,install("temperature"),ScalarReal(T));

  GetRNGstate();
  pdf = (double*)R_alloc(sizeof(double),ctx->n);
  for(i=0;i<INTEGER(count)[0];i++) {
    int    j,k,temp;
    double ru,cdf;
    SEXP order;
    
    PROTECT(order=allocVector(INTSXP,ctx->n));
    for(j=0;j<ctx->n;j++) INTEGER(order)[j] = j+1; /* Initialize order */
    REAL(scores)[i]  = 0.0;
    REAL(weights)[i] = 0.0;
    
    for(j=0;j<ctx->n;j++) {

/*       Rprintf("["); */
/*       for(k=0;k<j;k++) Rprintf("%d ",INTEGER(order)[k]); */
/*       Rprintf("\n"); */

      /* Calculate the PDF information */
      for(k=j;k<ctx->n;k++) {
	/*	Rprintf("\t%d",INTEGER(order)[k]);*/
	pdf[k] = invT*bnsl_score_extension(ctx,INTEGER(order),j,
					   INTEGER(order)[k]);
	/*	Rprintf(" %.2f\n",pdf[k]);*/
      }
      /* Sample an element */
      revsort(&pdf[j],&(INTEGER(order)[j]),ctx->n-j);
      norm = logsumexp(&pdf[j],ctx->n-j);
      ru   = unif_rand();cdf = 0.0;
      for(k=j;k<ctx->n;k++) {
	cdf += exp(pdf[k]-norm);
	if(ru < cdf) break;
      }
      /* Swap the true score into position */
      temp = INTEGER(order)[j];
      INTEGER(order)[j] = INTEGER(order)[k];
      INTEGER(order)[k] = temp;
      REAL(scores)[i]  += pdf[k];
      REAL(weights)[i] += pdf[k]-norm;
    }

    SET_VECTOR_ELT(orders,i,order);
    UNPROTECT(1);
  }
  PutRNGstate();

  bnsl_score_ctx_done(ctx);
  UNPROTECT(5);
  return ret;
}
