#include <Rdefines.h>
#include "common.h"
#include "bitstring.h"
#include <db.h>
#include "fastdb.h"
#include "fastdag.h"
#include <Rmath.h>


void fast_dag_sample_tempered(fast_dag_ctx ctx,double T,int max_iter,int callback(void*,char*,int,double),void*user); //fastdag.c

struct dag_info {
	DB*  db;
	DBT  key,data;
	int  n;
	int  unique;
	double epsilon,score,sampled,max_p;
};
void init_dag_info_(struct dag_info *di,int n,double score,double epsilon) {
	memset(di,0,sizeof(struct dag_info));
	db_create(&(di->db),NULL,0);
	di->db->open(di->db,NULL,NULL,NULL,DB_BTREE,DB_CREATE,0);
	di->n = n;
	di->score = score;
	di->sampled = R_NegInf;
	di->max_p   = 0;
	di->unique = 0;
	di->epsilon = epsilon;
}

int fast_dag_sample_(fast_dag_ctx ctx,int N,
	int callback(void*,char*,int,double),void*user) {
	int   i,size = bitstr_size(ctx->n)*ctx->n;
	char *dag  = R_alloc(sizeof(char),size);
	for(i=0;i<N;i++) {
		int j;
		double os = 1.0;
		double us = 0.0;
		for(j=0;j<ctx->n;j++) {
			double u = unif_rand();
			double c = 0.0;
			pset*  p = ctx->psets[j];
			if(!p) error("Unable to sample dag, no parent sets");
			//      Rprintf("(%.2f) ",u);
			while(p) { 
				//Rprintf("%.2f ",p->score);
				c += p->score;
				if(u <= c) break;
				p = p->next; 
			}
			//      Rprintf("[%.2f]\n",c);
			os *= p->score;
			us += p->real_score;
			memcpy(&dag[bitstr_size(ctx->n)*j],p->pset+p->n,bitstr_size(ctx->n));
		}
		if(!callback(user,dag,size,us)) break;
	}
	return i+1;
}

struct _x {
  int    count;
  double score;
};
int insert_dag_(void* userInfo,char*dag,int size,double score) {
	struct dag_info *di = (struct dag_info*)userInfo;
	di->key.data  = dag;
	di->key.size  = size;
	if(di->db->get(di->db,NULL,&(di->key),&(di->data),0) == 0) {
		struct _x *x = di->data.data;
		x->count++;
	} else {
		int i;
		di->unique++;
		struct _x x;
		x.count = 1;
		x.score = score;
		di->data.data = &x;
		di->data.size = sizeof(struct _x);
		double p = exp(score-di->sampled);
		if(di->sampled != R_NegInf)
			di->sampled  += log1p(p);
		else
			di->sampled = score;
		p = exp(score-di->score);
		if(di->max_p < p) di->max_p = p;
	}
	di->db->put(di->db,NULL,&(di->key),&(di->data),0);
/*	Rprintf("[%.2f %.2f]\n",di->sampled,di->score,di->epsilon);*/
	return (exp(di->sampled-di->score) < di->epsilon) ? 1 : 0;
}

SEXP BNSL_FastDagSampleWRP(SEXP D,SEXP os,SEXP K,SEXP N,SEXP E) {
	char          *vmax= vmaxget();
	fast_score_ctx ctx = fast_ctx(D,INTEGER(K)[0]);
	SEXP orders = GET_SLOT(os,install("orders"));
	SEXP scores = GET_SLOT(os,install("scores"));
	SEXP fracs,counts,max_p,sample_k;
	DB  *over;
	struct dag_info main;
	init_dag_info_(&main,ctx->n,0.0,0.0);
	GetRNGstate();
	Rprintf("%d unique orders\n",length(orders));
	int i,unique=0;
	PROTECT(fracs = allocVector(REALSXP,length(orders)));
	PROTECT(counts = allocVector(REALSXP,length(orders)));
	PROTECT(sample_k=allocVector(REALSXP,length(orders)));
	PROTECT(max_p=allocVector(REALSXP,length(orders)));
	for(i=0;i<length(orders);i++) {
		SEXP order = VECTOR_ELT(orders,i);
		struct dag_info di;
		init_dag_info_(&di,ctx->n,REAL(scores)[i] < 0 ? REAL(scores)[i]:-REAL(scores)[i],REAL(E)[0]);
		fast_dag_ctx dag = fast_dag(ctx,INTEGER(order),length(order));
		REAL(counts)[i] = fast_dag_sample_(dag,INTEGER(N)[0],insert_dag_,&di);
		REAL(fracs)[i] = exp(di.sampled-di.score);
		REAL(sample_k)[i] = di.unique;
		REAL(max_p)[i] = di.max_p;
		{
			DBC *cur;
			di.db->cursor(di.db,NULL,&cur,0);
			while(cur->c_get(cur,&(di.key),&(di.data),DB_NEXT)==0) {
				char *r = di.key.data;
				struct _x *c = di.data.data;
				main.key.data = r;
				main.key.size = di.key.size;
				if(main.db->get(main.db,NULL,&(main.key),&(main.data),0) != 0) {
					main.data.data= c;
					main.data.size = di.data.size;
					if(main.db->put(main.db,NULL,&(main.key),&(main.data),0)==0) unique++;
				}
			}
			cur->c_close(cur);
		}
		Rprintf(".");
		di.db->close(di.db,0);
		
	}
	Rprintf("\n%d unique dags\n",unique);
	{
		SEXP ret,scores,dags,sampled,overlap,vertices;
		PROTECT(ret = NEW_OBJECT(MAKE_CLASS("wrpDagSample")));
		PROTECT(scores  = allocVector(REALSXP,unique));
		PROTECT(dags    = allocVector(VECSXP,unique));

		DBC *cur;
		memset(&(main.key),0,sizeof(DBT));
		memset(&(main.data),0,sizeof(DBT));
		main.db->cursor(main.db,NULL,&cur,0);
		i=0;
		while(cur->c_get(cur,&(main.key),&(main.data),DB_NEXT)==0) {
			char      *r = main.key.data;
			struct _x *x = main.data.data;
			char      *c = (char*)R_alloc(sizeof(char),ctx->n);
			int        j,tmp;
			SEXP raw;

			PROTECT(raw = allocVector(RAWSXP,main.key.size));
			memcpy(RAW(raw),r,main.key.size);
			SET_VECTOR_ELT(dags,i,raw);
			UNPROTECT(1);
			REAL(scores)[i]  = (double)x->score;
			i++;
		}
		cur->c_close(cur);
		SET_SLOT(ret,install("dags"),dags);
		SET_SLOT(ret,install("p"),scores);
		SET_SLOT(ret,install("frac"),fracs);
		SET_SLOT(ret,install("n"),counts);
		SET_SLOT(ret,install("k"),sample_k);
		SET_SLOT(ret,install("p1"),max_p);
		UNPROTECT(7);
		main.db->close(main.db,0);
		PutRNGstate();
		vmaxset(vmax);
		return ret;
	}
}
